#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import pathlib
import subprocess
from textwrap import dedent
from unittest.mock import call

import pytest
from craft_providers import errors
from craft_providers.lxd import LXC, LXDError, lxc
from craft_providers.lxd.lxc import StdinType
from freezegun import freeze_time


@pytest.fixture()
def mock_getpid(mocker):
    return mocker.patch("os.getpid", return_value=123)


def test_lxc_run_default(mocker, tmp_path):
    """Test _lxc_run with default arguments."""
    mock_run = mocker.patch("subprocess.run")

    LXC()._run_lxc(command=["test-command"])

    mock_run.assert_called_once_with(
        ["lxc", "test-command"],
        check=True,
        stdin=subprocess.DEVNULL,
    )


@pytest.mark.parametrize("check", [True, False])
def test_lxc_run_with_check(check, mocker, tmp_path):
    """Test check parameter."""
    mock_run = mocker.patch("subprocess.run")

    LXC()._run_lxc(command=["test-command"], check=check, project="test-project")

    mock_run.assert_called_once_with(
        ["lxc", "--project", "test-project", "test-command"],
        check=check,
        stdin=subprocess.DEVNULL,
    )


def test_lxc_run_with_project(mocker, tmp_path):
    """Test _lxc_run with project."""
    mock_run = mocker.patch("subprocess.run")

    LXC()._run_lxc(command=["test-command"], project="test-project")

    mock_run.assert_called_once_with(
        ["lxc", "--project", "test-project", "test-command"],
        check=True,
        stdin=subprocess.DEVNULL,
    )


def test_lxc_run_with_stdin(mocker, tmp_path):
    """Test _lxc_run with stdin argument."""
    mock_run = mocker.patch("subprocess.run")

    LXC()._run_lxc(
        command=["test-command"], project="test-project", stdin=lxc.StdinType.NULL
    )

    mock_run.assert_called_once_with(
        ["lxc", "--project", "test-project", "test-command"],
        check=True,
        stdin=None,
    )


def test_lxc_run_with_input(mocker, tmp_path):
    """Test _lxc_run with input argument."""
    mock_run = mocker.patch("subprocess.run")

    LXC()._run_lxc(
        command=["test-command"],
        project="test-project",
        stdin=lxc.StdinType.NULL,
        input="test-input",
    )

    mock_run.assert_called_once_with(
        ["lxc", "--project", "test-project", "test-command"],
        check=True,
        input="test-input",
    )


def test_lxc_run_with_input_and_stdin(mocker, tmp_path):
    """`stdin` argument is ignored when `input` is passed."""
    mock_run = mocker.patch("subprocess.run")

    LXC()._run_lxc(
        command=["test-command"],
        project="test-project",
        stdin=lxc.StdinType.NULL,
        input="test-input",
    )

    mock_run.assert_called_once_with(
        ["lxc", "--project", "test-project", "test-command"],
        check=True,
        input="test-input",
    )


def test_config_device_add_disk(fake_process, tmp_path):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "device",
            "add",
            "test-remote:test-instance",
            "disk_foo",
            "disk",
            f"source={tmp_path.as_posix()}",
            "path=/mnt",
        ]
    )

    LXC().config_device_add_disk(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
        device="disk_foo",
        source=tmp_path,
        path=pathlib.PurePosixPath("/mnt"),
    )

    assert len(fake_process.calls) == 1


def test_config_device_add_disk_error(fake_process, tmp_path):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "device",
            "add",
            "test-remote:test-instance",
            "disk_foo",
            "disk",
            f"source={tmp_path.as_posix()}",
            "path=/mnt",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().config_device_add_disk(
            instance_name="test-instance",
            project="test-project",
            remote="test-remote",
            device="disk_foo",
            source=tmp_path,
            path=pathlib.PurePosixPath("/mnt"),
        )

    assert exc_info.value == LXDError(
        brief="Failed to add disk to instance 'test-instance'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_config_device_remove(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "device",
            "remove",
            "test-remote:test-instance",
            "device_foo",
        ]
    )

    LXC().config_device_remove(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
        device="device_foo",
    )

    assert len(fake_process.calls) == 1


def test_config_device_remove_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "device",
            "remove",
            "test-remote:test-instance",
            "device_foo",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().config_device_remove(
            instance_name="test-instance",
            project="test-project",
            remote="test-remote",
            device="device_foo",
        )

    assert exc_info.value == LXDError(
        brief="Failed to remove device from instance 'test-instance'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_config_device_show(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "device",
            "show",
            "test-remote:test-instance",
        ],
        stdout="test_mount:\n  path: /tmp\n  source: /tmp\n  type: disk\n",
    )

    devices = LXC().config_device_show(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1
    assert devices == {"test_mount": {"path": "/tmp", "source": "/tmp", "type": "disk"}}


def test_config_device_show_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "device",
            "show",
            "test-remote:test-instance",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().config_device_show(
            instance_name="test-instance",
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to show devices for instance 'test-instance'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_config_get(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "get",
            "test-remote:test-instance",
            "test-key",
        ],
    )

    LXC().config_get(
        instance_name="test-instance",
        key="test-key",
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1


def test_config_get_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "get",
            "test-remote:test-instance",
            "test-key",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().config_get(
            instance_name="test-instance",
            key="test-key",
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief=(
            "Failed to get value for config key 'test-key' "
            "for instance 'test-instance'."
        ),
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_config_set(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "set",
            "test-remote:test-instance",
            "test-key",
            "test-value",
        ],
    )

    LXC().config_set(
        instance_name="test-instance",
        key="test-key",
        value="test-value",
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1


def test_config_set_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "set",
            "test-remote:test-instance",
            "test-key",
            "test-value",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().config_set(
            instance_name="test-instance",
            key="test-key",
            value="test-value",
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief=(
            "Failed to set config key 'test-key' to 'test-value'"
            " for instance 'test-instance'."
        ),
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_copy(fake_process):
    """Test `copy()` with default arguments."""
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "default",
            "copy",
            "local:test-source-instance-name",
            "local:test-destination-instance-name",
        ],
    )

    LXC().copy(
        source_instance_name="test-source-instance-name",
        destination_instance_name="test-destination-instance-name",
    )

    assert len(fake_process.calls) == 1


def test_copy_all_opts(fake_process):
    """Test `copy() with all arguments defined`."""
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "copy",
            "test-source-remote:test-source-instance-name",
            "test-destination-remote:test-destination-instance-name",
        ],
    )

    LXC().copy(
        source_remote="test-source-remote",
        source_instance_name="test-source-instance-name",
        destination_remote="test-destination-remote",
        destination_instance_name="test-destination-instance-name",
        project="test-project",
    )

    assert len(fake_process.calls) == 1


def test_copy_error(fake_process):
    """A LXDError should be raised when the copy command fails."""
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "copy",
            "test-source-remote:test-source-instance-name",
            "test-destination-remote:test-destination-instance-name",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().copy(
            source_remote="test-source-remote",
            source_instance_name="test-source-instance-name",
            destination_remote="test-destination-remote",
            destination_instance_name="test-destination-instance-name",
            project="test-project",
        )

    assert exc_info.value == LXDError(
        brief=(
            "Failed to copy instance 'test-source-remote:test-source-instance-name' "
            "to 'test-destination-remote:test-destination-instance-name'."
        ),
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_delete(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "delete",
            "test-remote:test-instance",
        ],
    )

    LXC().delete(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1


def test_delete_all_opts(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "delete",
            "test-remote:test-instance",
            "--force",
        ],
    )

    LXC().delete(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
        force=True,
    )

    assert len(fake_process.calls) == 1


def test_delete_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "delete",
            "test-remote:test-instance",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().delete(
            instance_name="test-instance",
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to delete instance 'test-instance'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_exec(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "exec",
            "test-remote:test-instance",
            "--",
            "echo",
            "hi",
        ],
    )

    LXC().exec(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
        command=["echo", "hi"],
    )

    assert len(fake_process.calls) == 1


def test_exec_all_opts(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "exec",
            "test-remote:test-instance",
            "--cwd",
            "/tmp",
            "--mode",
            "interactive",
            "--",
            "echo",
            "hi",
        ],
    )

    LXC().exec(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
        command=["echo", "hi"],
        cwd="/tmp",
        mode="interactive",
        runner=subprocess.Popen,
    )

    assert len(fake_process.calls) == 1


def test_exec_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "exec",
            "test-remote:test-instance",
            "--",
            "/bin/false",
        ],
        returncode=1,
    )

    with pytest.raises(subprocess.CalledProcessError):
        LXC().exec(
            instance_name="test-instance",
            project="test-project",
            remote="test-remote",
            command=["/bin/false"],
            check=True,
        )


def test_file_pull(fake_process, tmp_path):
    source = pathlib.PurePosixPath("/root/foo")

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "file",
            "pull",
            "test-remote:test-instance/root/foo",
            tmp_path.as_posix(),
        ],
    )

    LXC().file_pull(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
        source=source,
        destination=tmp_path,
    )

    assert len(fake_process.calls) == 1


def test_file_pull_all_opts(fake_process, tmp_path):
    source = pathlib.PurePosixPath("/root/foo")

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "file",
            "pull",
            "test-remote:test-instance/root/foo",
            tmp_path.as_posix(),
            "--create-dirs",
            "--recursive",
        ],
    )

    LXC().file_pull(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
        source=source,
        destination=tmp_path,
        create_dirs=True,
        recursive=True,
    )

    assert len(fake_process.calls) == 1


def test_file_pull_error(fake_process, tmp_path):
    source = pathlib.PurePosixPath("/root/foo")

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "file",
            "pull",
            "test-remote:test-instance/root/foo",
            tmp_path.as_posix(),
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().file_pull(
            instance_name="test-instance",
            project="test-project",
            remote="test-remote",
            source=source,
            destination=tmp_path,
        )

    assert exc_info.value == LXDError(
        brief="Failed to pull file '/root/foo' from instance 'test-instance'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_file_push(fake_process, tmp_path):
    destination = pathlib.PurePosixPath("/root/foo")

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "file",
            "push",
            tmp_path.as_posix(),
            "test-remote:test-instance/root/foo",
        ],
    )

    LXC().file_push(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
        source=tmp_path,
        destination=destination,
    )

    assert len(fake_process.calls) == 1


def test_file_push_all_opts(fake_process, tmp_path):
    destination = pathlib.PurePosixPath("/root/foo")

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "file",
            "push",
            tmp_path.as_posix(),
            "test-remote:test-instance/root/foo",
            "--create-dirs",
            "--recursive",
            "--mode=0644",
            "--gid=1",
            "--uid=2",
        ],
    )

    LXC().file_push(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
        source=tmp_path,
        destination=destination,
        create_dirs=True,
        recursive=True,
        gid=1,
        uid=2,
        mode="0644",
    )

    assert len(fake_process.calls) == 1


def test_file_push_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "file",
            "push",
            "/somefile",
            "test-remote:test-instance/root/foo",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().file_push(
            instance_name="test-instance",
            project="test-project",
            remote="test-remote",
            source=pathlib.Path("/somefile"),
            destination=pathlib.PurePosixPath("/root/foo"),
        )

    assert exc_info.value == LXDError(
        brief="Failed to push file '/somefile' to instance 'test-instance'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_info(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "info",
            "test-remote:",
        ],
        stdout="config: {}\napi_extensions:\n - foo\n",
    )

    info = LXC().info(
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1
    assert info == {"config": {}, "api_extensions": ["foo"]}


def test_info_with_instance(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "info",
            "test-remote:test-instance",
        ],
        stdout="Name: test-instance\nLocation: none\n",
    )

    info = LXC().info(
        project="test-project",
        remote="test-remote",
        instance_name="test-instance",
    )

    assert len(fake_process.calls) == 1
    assert info == {"Name": "test-instance", "Location": "none"}


def test_info_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "info",
            "test-remote:",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().info(
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to get info for remote 'test-remote'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_info_parse_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "info",
            "test-remote:",
        ],
        stdout="fail:\nthis\n",
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().info(
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to parse lxc info.",
        details=(
            "* Command that failed: 'lxc --project test-project info test-remote:'\n"
            "* Command output: b'fail:\\nthis\\n'"
        ),
    )


@pytest.mark.usefixtures("mock_getpid")
def test_launch(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "launch",
            "test-image-remote:test-image",
            "test-remote:test-instance",
            "--config",
            "user.craft_providers.status=STARTING",
            "--config",
            "user.craft_providers.timer=2023-01-01T00:00:00+00:00",
            "--config",
            "user.craft_providers.pid=123",
        ]
    )

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "set",
            "test-remote:test-instance",
            "user.craft_providers.status",
            "PREPARING",
        ]
    )

    with freeze_time("2023-01-01"):
        LXC().launch(
            instance_name="test-instance",
            image="test-image",
            image_remote="test-image-remote",
            project="test-project",
            remote="test-remote",
        )

    assert len(fake_process.calls) == 1


@pytest.mark.usefixtures("mock_getpid")
def test_launch_failed_retry_check(fake_process, mocker):
    """Test that we use check_instance_status if launch fails."""
    mock_launch = mocker.patch("craft_providers.lxd.lxc.LXC._run_lxc")
    mock_launch.side_effect = [
        subprocess.CalledProcessError(1, ["lxc", "fail", " test"]),
    ]
    mock_check = mocker.patch("craft_providers.lxd.lxc.LXC.check_instance_status")
    mocker.patch("time.sleep")

    with freeze_time("2023-01-01"):
        LXC().launch(
            instance_name="test-instance",
            image="test-image",
            image_remote="test-image-remote",
            project="test-project",
            remote="test-remote",
        )

    assert mock_launch.mock_calls == [
        call(
            [
                "launch",
                "test-image-remote:test-image",
                "test-remote:test-instance",
                "--config",
                "user.craft_providers.status=STARTING",
                "--config",
                "user.craft_providers.timer=2023-01-01T00:00:00+00:00",
                "--config",
                "user.craft_providers.pid=123",
            ],
            capture_output=True,
            stdin=StdinType.INTERACTIVE,
            project="test-project",
        ),
    ]

    assert mock_check.mock_calls == [
        call(
            instance_name="test-instance", project="test-project", remote="test-remote"
        )
    ]


@pytest.mark.usefixtures("mock_getpid")
def test_launch_failed_retry_failed(fake_process, mocker):
    """Test that we retry launching an instance if it fails, but failed more than 3 times."""
    mock_launch = mocker.patch("craft_providers.lxd.lxc.LXC._run_lxc")
    mock_launch.side_effect = subprocess.CalledProcessError(1, ["lxc", "fail", " test"])
    mock_check = mocker.patch("craft_providers.lxd.lxc.LXC.check_instance_status")
    mock_check.side_effect = LXDError("test")
    mocker.patch("time.sleep")

    with pytest.raises(LXDError):
        with freeze_time("2023-01-01"):
            LXC().launch(
                instance_name="test-instance",
                image="test-image",
                image_remote="test-image-remote",
                project="test-project",
                remote="test-remote",
            )

    assert mock_launch.mock_calls == [
        call(
            [
                "launch",
                "test-image-remote:test-image",
                "test-remote:test-instance",
                "--config",
                "user.craft_providers.status=STARTING",
                "--config",
                "user.craft_providers.timer=2023-01-01T00:00:00+00:00",
                "--config",
                "user.craft_providers.pid=123",
            ],
            capture_output=True,
            stdin=StdinType.INTERACTIVE,
            project="test-project",
        ),
        call(
            ["delete", "test-remote:test-instance", "--force"],
            capture_output=True,
            project="test-project",
        ),
        call(
            [
                "launch",
                "test-image-remote:test-image",
                "test-remote:test-instance",
                "--config",
                "user.craft_providers.status=STARTING",
                "--config",
                "user.craft_providers.timer=2023-01-01T00:00:00+00:00",
                "--config",
                "user.craft_providers.pid=123",
            ],
            capture_output=True,
            stdin=StdinType.INTERACTIVE,
            project="test-project",
        ),
        call(
            ["delete", "test-remote:test-instance", "--force"],
            capture_output=True,
            project="test-project",
        ),
        call(
            [
                "launch",
                "test-image-remote:test-image",
                "test-remote:test-instance",
                "--config",
                "user.craft_providers.status=STARTING",
                "--config",
                "user.craft_providers.timer=2023-01-01T00:00:00+00:00",
                "--config",
                "user.craft_providers.pid=123",
            ],
            capture_output=True,
            stdin=StdinType.INTERACTIVE,
            project="test-project",
        ),
    ]


@pytest.mark.usefixtures("mock_getpid")
def test_launch_all_opts(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "launch",
            "test-image-remote:test-image",
            "test-remote:test-instance",
            "--ephemeral",
            "--config",
            "test-key=test-value",
            "--config",
            "test-key2=test-value2",
            "--config",
            "user.craft_providers.status=STARTING",
            "--config",
            "user.craft_providers.timer=2023-01-01T00:00:00+00:00",
            "--config",
            "user.craft_providers.pid=123",
        ]
    )

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "set",
            "test-remote:test-instance",
            "user.craft_providers.status",
            "PREPARING",
        ]
    )

    with freeze_time("2023-01-01"):
        LXC().launch(
            instance_name="test-instance",
            image="test-image",
            image_remote="test-image-remote",
            project="test-project",
            remote="test-remote",
            config_keys={"test-key": "test-value", "test-key2": "test-value2"},
            ephemeral=True,
        )

    assert len(fake_process.calls) == 1


@pytest.mark.usefixtures("mock_getpid")
def test_launch_error(fake_process, mocker):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "launch",
            "test-image-remote:test-image",
            "test-remote:test-instance",
            "--config",
            "user.craft_providers.status=STARTING",
            "--config",
            "user.craft_providers.timer=2023-01-01T00:00:00+00:00",
            "--config",
            "user.craft_providers.pid=123",
        ],
        returncode=1,
        occurrences=4,
    )

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "delete",
            "test-remote:test-instance",
            "--force",
        ],
        returncode=0,
        occurrences=4,
    )

    mocker.patch(
        "craft_providers.lxd.lxc.LXC.check_instance_status"
    ).side_effect = LXDError("Failed to get instance status.")

    mocker.patch("time.sleep")

    with pytest.raises(LXDError) as exc_info:
        with freeze_time("2023-01-01"):
            LXC().launch(
                instance_name="test-instance",
                image="test-image",
                image_remote="test-image-remote",
                project="test-project",
                remote="test-remote",
            )

    assert exc_info.value == LXDError(
        brief="Failed to launch instance 'test-instance'.",
        details="* Command that failed: 'lxc --project test-project launch test-image-remote:test-image test-remote:test-instance --config user.craft_providers.status=STARTING --config user.craft_providers.timer=2023-01-01T00:00:00+00:00 --config user.craft_providers.pid=123'\n* Command exit code: 1",
        resolution=None,
    )


def test_check_instance_status(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "info",
            "test-remote:test-instance",
        ],
        returncode=0,
        stdout=dedent(
            """\
            Name: test-instance
            Status: STOPPED
            Type: container
            Architecture: x86_64
            Created: 2023/08/02 14:04 EDT
            Last Used: 2023/08/08 09:44 EDT
            """
        ),
    )

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "get",
            "test-remote:test-instance",
            "user.craft_providers.status",
        ],
        returncode=0,
        stdout="FINISHED",
    )

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "get",
            "test-remote:test-instance",
            "user.craft_providers.timer",
        ],
        returncode=0,
        stdout="2023-01-01T00:00:00+00:00",
    )

    LXC().check_instance_status(
        instance_name="test-instance", project="test-project", remote="test-remote"
    )


def test_check_instance_status_retry(fake_process, mocker):
    time_time = mocker.patch("time.time")
    time_time.side_effect = [0, 10]
    mocker.patch("time.sleep")
    mock_instance = mocker.patch("craft_providers.lxd.lxc.LXC.info")
    mock_instance.side_effect = [
        {"Status": "STOPPED"},
        {"Status": "STOPPED"},
    ]
    mock_instance_config = mocker.patch("craft_providers.lxd.lxc.LXC.config_get")
    mock_instance_config.side_effect = [
        "STARTING",
        "10",
        "FINISHED",
        "20",
    ]

    LXC().check_instance_status(
        instance_name="test-instance", project="test-project", remote="test-remote"
    )


def test_check_instance_status_boot_failed(fake_process, mocker):
    time_time = mocker.patch("time.time")
    time_time.return_value = 0
    mocker.patch("time.sleep")
    mock_instance = mocker.patch("craft_providers.lxd.lxc.LXC.info")
    mock_instance.return_value = {"Status": "STOPPED"}
    mock_instance_config = mocker.patch("craft_providers.lxd.lxc.LXC.config_get")
    mock_instance_config.side_effect = [
        "STARTING",
        "2023-01-01T00:00:00+00:00",
    ] * 20

    with pytest.raises(LXDError):
        LXC().check_instance_status(
            instance_name="test-instance", project="test-project", remote="test-remote"
        )


def test_check_instance_status_wait(fake_process, mocker):
    time_time = mocker.patch("time.time")
    time_time.return_value = 0
    mocker.patch("time.sleep")
    mock_instance = mocker.patch("craft_providers.lxd.lxc.LXC.info")
    mock_instance.side_effect = [
        {"Status": "STOPPED"},  # STARTING
        {"Status": "RUNNING"},  # STARTING
        {"Status": "RUNNING"},  # PREPARING
        {"Status": "RUNNING"},  # PREPARING
        {"Status": "RUNNING"},  # FINISHED
        {"Status": "STOPPED"},  # FINISHED
    ]
    mock_instance_config = mocker.patch("craft_providers.lxd.lxc.LXC.config_get")
    mock_instance_config.side_effect = [
        "STARTING",
        "2023-01-01T00:00:00+00:00",
        "STARTING",
        "2023-01-01T00:00:05+00:00",
        "PREPARING",
        "2023-01-01T00:00:10+00:00",
        "PREPARING",
        "2023-01-01T00:00:15+00:00",
        "FINISHED",
        "2023-01-01T00:00:20+00:00",
        "FINISHED",
        "2023-01-01T00:00:30+00:00",
    ]

    LXC().check_instance_status(
        instance_name="test-instance", project="test-project", remote="test-remote"
    )
    assert mock_instance.call_count == 6
    assert mock_instance_config.call_count == 12


def test_check_instance_status_lxd_error(fake_process, mocker):
    time_time = mocker.patch("time.time")
    time_time.side_effect = [0, 1000, 2000]
    mocker.patch("time.sleep")

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "info",
            "test-remote:test-instance",
        ],
        returncode=0,
        stdout=dedent(
            """\
            Name: test-instance
            Status: STOPPED
            Type: container
            Architecture: x86_64
            Created: 2023/08/02 14:04 EDT
            Last Used: 2023/08/08 09:44 EDT
            """
        ),
        occurrences=1000,
    )

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "get",
            "test-remote:test-instance",
            "user.craft_providers.status",
        ],
        returncode=1,
        stdout="",
        occurrences=1000,
    )

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "get",
            "test-remote:test-instance",
            "user.craft_providers.timer",
        ],
        returncode=1,
        stdout="",
        occurrences=1000,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().check_instance_status(
            instance_name="test-instance", project="test-project", remote="test-remote"
        )

    assert exc_info.value == LXDError(
        brief="Failed to get value for config key 'user.craft_providers.status' for instance 'test-instance'.",
        details="* Command that failed: 'lxc --project test-project config get test-remote:test-instance user.craft_providers.status'\n* Command exit code: 1",
        resolution=None,
    )


def test_check_instance_status_lxd_error_retry(fake_process, mocker):
    time_time = mocker.patch("time.time")
    time_time.side_effect = [0, 100, 200, 300, 400, 500, 600, 700, 800, 900]
    mocker.patch("time.sleep")
    mock_instance = mocker.patch("craft_providers.lxd.lxc.LXC.info")
    mock_instance.side_effect = [
        LXDError(
            brief="Failed to get instance info.",
            details="* Command that failed: 'lxc --project test-project info test-remote:test-instance'\n* Command exit code: 1",
            resolution=None,
        ),
        {"Status": "STOPPED"},
        {"Status": "RUNNING"},
        {"Status": "RUNNING"},
        LXDError(
            brief="Failed to get instance info.",
            details="* Command that failed: 'lxc --project test-project info test-remote:test-instance'\n* Command exit code: 1",
            resolution=None,
        ),
        {"Status": "RUNNING"},
        {"Status": "RUNNING"},
        {"Status": "RUNNING"},
        {"Status": "STOPPED"},
    ]
    mock_instance_config = mocker.patch("craft_providers.lxd.lxc.LXC.config_get")
    mock_instance_config.side_effect = [
        LXDError(
            brief="Failed to get instance info.",
            details="* Command that failed: 'lxc --project test-project info test-remote:test-instance'\n* Command exit code: 1",
            resolution=None,
        ),
        LXDError(
            brief="Failed to get instance info.",
            details="* Command that failed: 'lxc --project test-project info test-remote:test-instance'\n* Command exit code: 1",
            resolution=None,
        ),
        "STARTING",
        "2023-01-01T00:00:00+00:00",
        "STARTING",
        "2023-01-01T00:00:10+00:00",
        "PREPARING",
        "2023-01-01T00:00:20+00:00",
        LXDError(
            brief="Failed to get instance info.",
            details="* Command that failed: 'lxc --project test-project info test-remote:test-instance'\n* Command exit code: 1",
            resolution=None,
        ),
        "FINISHED",
        "2023-01-01T00:00:40+00:00",
        "FINISHED",
        "2023-01-01T00:00:50+00:00",
    ]

    LXC().check_instance_status(
        instance_name="test-instance", project="test-project", remote="test-remote"
    )


def test_check_instance_status_error_timeout(fake_process, mocker):
    time_time = mocker.patch("time.time")
    time_time.side_effect = [0, 1000, 2000]
    mocker.patch("time.sleep")

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "info",
            "test-remote:test-instance",
        ],
        returncode=0,
        stdout=dedent(
            """\
            Name: test-instance
            Status: STOPPED
            Type: container
            Architecture: x86_64
            Created: 2023/08/02 14:04 EDT
            Last Used: 2023/08/08 09:44 EDT
            """
        ),
        occurrences=1000,
    )

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "get",
            "test-remote:test-instance",
            "user.craft_providers.status",
        ],
        returncode=0,
        stdout="STARTING",
        occurrences=1000,
    )

    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "get",
            "test-remote:test-instance",
            "user.craft_providers.timer",
        ],
        returncode=0,
        stdout="2023-01-01T00:00:00+00:00",
        occurrences=1000,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().check_instance_status(
            instance_name="test-instance",
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Timed out waiting for instance to be ready."
    )


def test_has_image(fake_process):
    lxc = LXC()
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "image",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        stdout="- aliases:\n  - name: image1\n- aliases:\n  - name: image2\n",
        occurrences=3,
    )
    fake_process.keep_last_process(True)

    assert lxc.has_image("image1", project="test-project", remote="test-remote") is True
    assert lxc.has_image("image2", project="test-project", remote="test-remote") is True
    assert (
        lxc.has_image("image3", project="test-project", remote="test-remote") is False
    )


def test_image_copy(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "image",
            "copy",
            "test-image-remote:test-image",
            "test-remote:",
        ]
    )

    LXC().image_copy(
        image="test-image",
        image_remote="test-image-remote",
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1


def test_image_copy_all_opts(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "image",
            "copy",
            "test-image-remote:test-image",
            "test-remote:",
            "--alias=test-alias",
        ]
    )

    LXC().image_copy(
        image="test-image",
        image_remote="test-image-remote",
        alias="test-alias",
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1


def test_image_copy_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "image",
            "copy",
            "test-image-remote:test-image",
            "test-remote:",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().image_copy(
            image="test-image",
            image_remote="test-image-remote",
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to copy image 'test-image'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_image_delete(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "image",
            "delete",
            "test-remote:test-image",
        ]
    )

    LXC().image_delete(
        image="test-image",
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1


def test_image_delete_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "image",
            "delete",
            "test-remote:test-image",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().image_delete(
            image="test-image",
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to delete image 'test-image'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_image_list(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "image",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        stdout="- image1: stuff\n- image2: stuff\n",
    )

    images = LXC().image_list(
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1
    assert images == [{"image1": "stuff"}, {"image2": "stuff"}]


def test_image_list_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "image",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().image_list(
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to list images for project 'test-project'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_image_list_parse_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "image",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        stdout="fail:\nthis\n",
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().image_list(
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to parse lxc image list.",
        details=(
            "* Command that failed:"
            " 'lxc --project test-project image list test-remote: --format=yaml'\n"
            "* Command output: b'fail:\\nthis\\n'"
        ),
    )


def test_list(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        stdout="- name: test1\n- name: test2\n",
    )

    container_names = LXC().list(
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1
    assert container_names == [{"name": "test1"}, {"name": "test2"}]


def test_list_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().list(
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to list instances for project 'test-project'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_list_parse_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        stdout="fail:\nthis\n",
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().list(
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to parse lxc list.",
        details=(
            "* Command that failed:"
            " 'lxc --project test-project list test-remote: --format=yaml'\n"
            "* Command output: b'fail:\\nthis\\n'"
        ),
    )


def test_list_names(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        stdout="- name: test1\n- name: test2\n",
    )

    container_names = LXC().list_names(
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1
    assert container_names == ["test1", "test2"]


def test_list_names_parse_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        stdout="- foo: bar",
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().list_names(
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to parse lxc list.",
        details=("* Data received from lxc list: [{'foo': 'bar'}]"),
    )


def test_profile_edit(fake_process):
    stdin_records = []
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "profile",
            "edit",
            "test-remote:test-profile",
        ],
        stdin_callable=stdin_records.append,
        stdout="- name: test1\n- name: test2\n",
    )

    LXC().profile_edit(
        profile="test-profile",
        config={"test-key": "test-value", "test-key2": "test-value2"},
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1
    assert stdin_records == [b"test-key: test-value\ntest-key2: test-value2\n"]


def test_profile_edit_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "profile",
            "edit",
            "test-remote:test-profile",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().profile_edit(
            profile="test-profile",
            config={"test-key": "test-value", "test-key2": "test-value2"},
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to set profile 'test-profile'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_profile_show(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "profile",
            "show",
            "test-remote:test-profile",
        ],
        stdout="config: {}\ndescription: LXD profile\n",
    )

    profile = LXC().profile_show(
        profile="test-profile",
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1
    assert profile == {"config": {}, "description": "LXD profile"}


def test_profile_show_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "profile",
            "show",
            "test-remote:test-profile",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().profile_show(
            profile="test-profile",
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to show profile 'test-profile'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_project_create(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "project",
            "create",
            "test-remote:test-project",
        ],
    )

    LXC().project_create(
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1


def test_project_create_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "project",
            "create",
            "test-remote:test-project",
        ],
        returncode=1,
    )
    fake_process.register_subprocess(
        [
            "lxc",
            "project",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        stdout="- config:\n  name: default\n",
        returncode=0,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().project_create(
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to create project 'test-project'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_project_create_error_race(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "project",
            "create",
            "test-remote:test-project",
        ],
        returncode=1,
    )
    fake_process.register_subprocess(
        [
            "lxc",
            "project",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        stdout="- config:\n  name: default\n- config:\n  name: test-project\n",
        returncode=0,
    )

    LXC().project_create(
        project="test-project",
        remote="test-remote",
    )
    assert len(fake_process.calls) == 2


def test_project_delete(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "project",
            "delete",
            "test-remote:test-project",
        ],
    )

    LXC().project_delete(
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1


def test_project_delete_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "project",
            "delete",
            "test-remote:test-project",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().project_delete(
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to delete project 'test-project'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_project_list(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "project",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        stdout="- name: default\n- name: myproject\n",
    )

    projects = LXC().project_list(
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1
    assert projects == ["default", "myproject"]


def test_project_list_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "project",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().project_list(
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to list projects on remote 'test-remote'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_project_list_parse_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "project",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        stdout="fail:\nthis\n",
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().project_list(
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to parse lxc project list.",
        details=(
            "* Command that failed: 'lxc project list test-remote: --format=yaml'\n"
            "* Command output: b'fail:\\nthis\\n'"
        ),
    )


def test_project_list_parse_error_missing_name(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "project",
            "list",
            "test-remote:",
            "--format=yaml",
        ],
        stdout="- foo: bar",
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().project_list(
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to parse lxc project list.",
        details=(
            "* Command that failed: 'lxc project list test-remote: --format=yaml'\n"
            "* Command output: b'- foo: bar'"
        ),
    )


def test_publish(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "publish",
            "test-remote:test-instance",
            "test-image-remote:",
        ],
    )

    LXC().publish(
        image_remote="test-image-remote",
        instance_name="test-instance",
        remote="test-remote",
        project="test-project",
    )

    assert len(fake_process.calls) == 1


def test_publish_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "publish",
            "test-remote:test-instance",
            "test-image-remote:",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().publish(
            instance_name="test-instance",
            remote="test-remote",
            image_remote="test-image-remote",
            project="test-project",
        )

    assert exc_info.value == LXDError(
        brief="Failed to publish image from 'test-instance'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_publish_all_opts(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "publish",
            "test-remote:test-instance",
            "test-image-remote:",
            "--alias=test-alias",
            "--force",
        ],
    )

    LXC().publish(
        alias="test-alias",
        force=True,
        image_remote="test-image-remote",
        instance_name="test-instance",
        remote="test-remote",
        project="test-project",
    )

    assert len(fake_process.calls) == 1


def test_remote_add(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "remote",
            "add",
            "test-remote",
            "test-remote-addr",
            "--protocol=test-protocol",
        ],
    )

    LXC().remote_add(
        remote="test-remote",
        addr="test-remote-addr",
        protocol="test-protocol",
    )

    assert len(fake_process.calls) == 1


def test_remote_add_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "remote",
            "add",
            "test-remote",
            "test-remote-addr",
            "--protocol=test-protocol",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().remote_add(
            remote="test-remote",
            addr="test-remote-addr",
            protocol="test-protocol",
        )

    assert exc_info.value == LXDError(
        brief="Failed to add remote 'test-remote'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_remote_list(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "remote",
            "list",
            "--format=yaml",
        ],
        stdout="r1:\n  addr: a1\nr2:\n  addr: a2\n",
    )

    remotes = LXC().remote_list()

    assert len(fake_process.calls) == 1
    assert remotes == {"r1": {"addr": "a1"}, "r2": {"addr": "a2"}}


def test_remote_list_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "remote",
            "list",
            "--format=yaml",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().remote_list()

    assert exc_info.value == LXDError(
        brief="Failed to list remotes.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_remote_list_parse_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "remote",
            "list",
            "--format=yaml",
        ],
        stdout=b"fail:\nthis",
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().remote_list()

    assert exc_info.value == LXDError(
        brief="Failed to parse lxc remote list.",
        details=(
            "* Command that failed: 'lxc remote list --format=yaml'\n"
            "* Command output: b'fail:\\nthis'"
        ),
    )


def test_start(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "start",
            "test-remote:test-instance",
        ],
    )

    LXC().start(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1


def test_start_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "start",
            "test-remote:test-instance",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().start(
            instance_name="test-instance",
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to start 'test-instance'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_restart(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "restart",
            "test-remote:test-instance",
        ],
    )

    LXC().restart(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1


def test_restart_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "restart",
            "test-remote:test-instance",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().restart(
            instance_name="test-instance",
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to restart 'test-instance'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_stop(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "stop",
            "test-remote:test-instance",
        ],
    )

    LXC().stop(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
    )

    assert len(fake_process.calls) == 1


def test_stop_all_opts(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "stop",
            "test-remote:test-instance",
            "--force",
            "--timeout=4",
        ],
    )

    LXC().stop(
        instance_name="test-instance",
        project="test-project",
        remote="test-remote",
        force=True,
        timeout=4,
    )

    assert len(fake_process.calls) == 1


def test_stop_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "stop",
            "test-remote:test-instance",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXC().stop(
            instance_name="test-instance",
            project="test-project",
            remote="test-remote",
        )

    assert exc_info.value == LXDError(
        brief="Failed to stop 'test-instance'.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_yaml_loader_invalid_timestamp():
    # This would throw a `ValueError: year 0 is out of range` if loader is
    # resolving timestamps.
    data = "last_used_at: 0000-01-01T00:00:00Z"

    obj = lxc.load_yaml(data)

    assert "last_used_at" in obj
    assert isinstance(obj["last_used_at"], str)



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Fixtures for LXD integration tests."""
import os
import random
import string
import subprocess
import time
from contextlib import contextmanager
from typing import Any, Dict, Optional

import pytest
from craft_providers.lxd import LXC
from craft_providers.lxd import project as lxc_project
from craft_providers.lxd.lxd_instance import LXDInstance
from craft_providers.lxd.lxd_provider import LXDProvider


@pytest.fixture(autouse=True, scope="module")
def installed_lxd_required(installed_lxd):
    """All LXD integration tests required LXD to be installed."""


@pytest.fixture()
@pytest.mark.with_sudo()
def installed_lxd_without_init(uninstalled_lxd):
    """Ensure lxd is installed, but not initialized.

    Initialize LXD on cleanup to ensure it is functional for other tests.

    Requires CRAFT_PROVIDERS_TESTS_ENABLE_LXD_INSTALL=1.
    """
    if os.environ.get("CRAFT_PROVIDERS_TESTS_ENABLE_LXD_INSTALL") != "1":
        pytest.skip("lxd not installed, skipped")
    subprocess.run(["sudo", "snap", "install", "lxd"], check=True)
    subprocess.run(["sudo", "lxd", "waitready"], check=True)
    yield
    subprocess.run(["sudo", "lxd", "init", "--auto"], check=True)


@contextmanager
def tmp_instance(
    *,
    name: str,
    config_keys: Optional[Dict[str, Any]] = None,
    ephemeral: bool = True,
    image: str = "22.04",
    image_remote: str = "ubuntu",
    project: str,
    remote: str = "local",
    lxc: LXC = LXC(),
):
    if config_keys is None:
        config_keys = {}

    instance = LXDInstance(
        name=name,
        project=project,
        remote=remote,
        lxc=lxc,
    )
    instance_name = instance.instance_name

    lxc.launch(
        instance_name=instance_name,
        config_keys=config_keys,
        ephemeral=ephemeral,
        image=image,
        image_remote=image_remote,
        project=project,
        remote=remote,
    )

    # Make sure container is ready
    for _ in range(0, 60):
        proc = lxc.exec(
            command=["systemctl", "is-system-running"],
            instance_name=instance_name,
            project=project,
            remote=remote,
            capture_output=True,
            check=False,
            text=True,
        )

        running_state = proc.stdout.strip()
        if running_state in ["running", "degraded"]:
            break

        time.sleep(0.1)

    yield instance_name

    if instance_name in lxc.list_names(project=project, remote=remote):
        lxc.delete(
            instance_name=instance_name, project=project, remote=remote, force=True
        )

    assert instance_name not in lxc.list_names(project=project, remote=remote)


@pytest.fixture()
def lxc():
    return LXC()


@pytest.fixture()
def project_name():
    """Create temporary LXD project and assert expected properties."""
    return "ptest-" + "".join(random.choices(string.ascii_uppercase, k=4))


@pytest.fixture()
def project(lxc, project_name):
    """Create temporary LXD project and assert expected properties."""
    lxc_project.create_with_default_profile(lxc=lxc, project=project_name)

    projects = lxc.project_list()
    assert project_name in projects

    instances = lxc.list(project=project_name)
    assert instances == []

    expected_cfg = lxc.profile_show(profile="default", project="default")
    expected_cfg["used_by"] = []

    assert lxc.profile_show(profile="default", project=project_name) == expected_cfg

    yield project_name

    lxc_project.purge(lxc=lxc, project=project_name)


@pytest.fixture(scope="session")
def session_project(installed_lxd):
    lxc = LXC()
    project_name = "craft-providers-test-session"
    lxc_project.create_with_default_profile(lxc=lxc, project=project_name)

    projects = lxc.project_list()
    assert project_name in projects

    instances = lxc.list(project=project_name)
    assert instances == []

    expected_cfg = lxc.profile_show(profile="default", project="default")
    expected_cfg["used_by"] = []

    assert lxc.profile_show(profile="default", project=project_name) == expected_cfg

    yield project_name

    lxc_project.purge(lxc=lxc, project=project_name)


@pytest.fixture(scope="session")
def session_provider(session_project):
    return LXDProvider(lxd_project=session_project)



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import pytest
from craft_providers import errors
from craft_providers.lxd import LXD, LXDError


def test_init(fake_process):
    fake_process.register_subprocess(
        [
            "lxd",
            "init",
        ]
    )

    LXD().init()

    assert len(fake_process.calls) == 1


def test_init_all_opts(fake_process):
    fake_process.register_subprocess(
        [
            "sudo",
            "lxd",
            "init",
            "--auto",
        ]
    )

    LXD().init(
        auto=True,
        sudo=True,
    )

    assert len(fake_process.calls) == 1


def test_init_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxd",
            "init",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXD().init()

    assert exc_info.value == LXDError(
        brief="Failed to init LXD.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


@pytest.mark.parametrize(
    ("version", "compatible"),
    [
        ("3.0", False),
        ("3.1.4", False),
        ("3.10", False),
        ("4.0", True),
        ("4.1.4", True),
        ("4.10", True),
    ],
)
def test_is_supported_version(fake_process, version, compatible):
    fake_process.register_subprocess(
        [
            "lxd",
            "version",
        ],
        stdout=version,
    )

    assert LXD().is_supported_version() == compatible
    assert len(fake_process.calls) == 1


@pytest.mark.parametrize("version_data", ["", "invalid"])
def test_is_supported_version_parse_error(fake_process, version_data):
    fake_process.register_subprocess(
        [
            "lxd",
            "version",
        ],
        stdout=version_data,
    )

    with pytest.raises(LXDError) as exc_info:
        LXD().is_supported_version()

    assert exc_info.value == LXDError(
        brief="Failed to parse LXD version.",
        details=f"Version data returned: {version_data!r}",
    )


def test_version(fake_process):
    fake_process.register_subprocess(
        [
            "lxd",
            "version",
        ],
        stdout="test-version",
    )

    version = LXD().version()

    assert len(fake_process.calls) == 1
    assert version == "test-version"


def test_version_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxd",
            "version",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXD().version()

    assert exc_info.value == LXDError(
        brief="Failed to query LXD version.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_wait_ready(fake_process):
    fake_process.register_subprocess(
        [
            "lxd",
            "waitready",
        ]
    )

    LXD().wait_ready()

    assert len(fake_process.calls) == 1


def test_wait_ready_all_opts(fake_process):
    fake_process.register_subprocess(
        [
            "sudo",
            "lxd",
            "waitready",
            "--timeout=5",
        ]
    )

    LXD().wait_ready(
        sudo=True,
        timeout=5,
    )

    assert len(fake_process.calls) == 1


def test_wait_ready_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxd",
            "waitready",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError) as exc_info:
        LXD().wait_ready()

    assert exc_info.value == LXDError(
        brief="Failed to wait for LXD to get ready.",
        details=errors.details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import os
import shutil
import sys
from unittest import mock

import pytest
from craft_providers.lxd import (
    LXC,
    LXD,
    LXDError,
    LXDInstallationError,
    install,
    installer,
    is_initialized,
    is_installed,
    is_user_permitted,
)


@pytest.fixture()
def mock_is_installed():
    with mock.patch(
        "craft_providers.lxd.installer.is_installed", return_value=True
    ) as mock_is_installed:
        yield mock_is_installed


@pytest.fixture()
def mock_is_user_permitted():
    with mock.patch(
        "craft_providers.lxd.installer.is_user_permitted", return_value=True
    ) as mock_is_user_permitted:
        yield mock_is_user_permitted


@pytest.fixture()
def mock_is_initialized():
    with mock.patch(
        "craft_providers.lxd.installer.is_initialized", return_value=True
    ) as mock_is_initialized:
        yield mock_is_initialized


@pytest.fixture()
def mock_lxd():
    mock_lxd = mock.Mock(spec=LXD)
    mock_lxd.is_supported_version.return_value = True
    mock_lxd.version.return_value = "4.4"
    mock_lxd.minimum_required_version = LXD.minimum_required_version

    return mock_lxd


@pytest.fixture()
def mock_os_geteuid():
    with mock.patch.object(
        os, "geteuid", return_value=500, create=True
    ) as mock_os_geteuid:
        yield mock_os_geteuid


@pytest.fixture()
def mock_os_getgroups():
    with mock.patch.object(
        os, "getgroups", return_value=[], create=True
    ) as mock_os_getgroups:
        yield mock_os_getgroups


@pytest.fixture()
def mock_os_access():
    with mock.patch.object(
        os, "access", return_value=True, create=True
    ) as mock_os_access:
        yield mock_os_access


@pytest.mark.parametrize("platform", ["win32", "darwin", "other"])
def test_install_unsupported_platform(mocker, platform):
    mocker.patch.object(sys, "platform", platform)

    with pytest.raises(LXDInstallationError) as exc_info:
        install()

    assert exc_info.value == LXDInstallationError(
        f"unsupported platform {sys.platform!r}"
    )


def test_install_without_sudo(
    fake_process, mocker, mock_is_user_permitted, mock_os_geteuid
):
    mocker.patch.object(sys, "platform", "linux")
    mock_os_geteuid.return_value = 0
    fake_process.register_subprocess(
        [
            "snap",
            "install",
            "lxd",
        ]
    )
    fake_process.register_subprocess(
        [
            "lxd",
            "waitready",
        ]
    )
    fake_process.register_subprocess(
        [
            "lxd",
            "init",
            "--auto",
        ]
    )
    fake_process.register_subprocess(
        [
            "lxd",
            "version",
        ],
        stdout="4.0",
    )

    version = install(sudo=False)

    assert version == "4.0"
    mock_is_user_permitted.assert_called_once()


def test_install_with_sudo(
    fake_process, mocker, mock_is_user_permitted, mock_os_geteuid
):
    mocker.patch.object(sys, "platform", "linux")
    fake_process.register_subprocess(
        [
            "sudo",
            "snap",
            "install",
            "lxd",
        ]
    )
    fake_process.register_subprocess(
        [
            "sudo",
            "lxd",
            "waitready",
        ]
    )
    fake_process.register_subprocess(
        [
            "sudo",
            "lxd",
            "init",
            "--auto",
        ]
    )
    fake_process.register_subprocess(
        [
            "lxd",
            "version",
        ],
        stdout="4.0",
    )

    version = install(sudo=True)

    assert version == "4.0"
    mock_is_user_permitted.assert_called_once()


def test_install_requires_sudo(mocker, mock_is_user_permitted, mock_os_geteuid):
    mocker.patch.object(sys, "platform", "linux")

    with pytest.raises(LXDInstallationError) as exc_info:
        install(sudo=False)

    assert exc_info.value == LXDInstallationError(
        "sudo required if not running as root"
    )


@pytest.mark.skipif(sys.platform != "linux", reason=f"unsupported on {sys.platform}")
def test_install_requires_user_to_be_added_to_lxd_group(
    fake_process, mock_is_user_permitted
):
    mock_is_user_permitted.return_value = False
    fake_process.register_subprocess(
        [
            "sudo",
            "snap",
            "install",
            "lxd",
        ]
    )
    fake_process.register_subprocess(
        [
            "sudo",
            "lxd",
            "waitready",
        ]
    )
    fake_process.register_subprocess(
        [
            "sudo",
            "lxd",
            "init",
            "--auto",
        ]
    )
    fake_process.register_subprocess(
        [
            "lxd",
            "version",
        ],
        stdout="4.0",
    )

    with pytest.raises(LXDInstallationError) as exc_info:
        install(sudo=True)

    assert exc_info.value == LXDInstallationError(
        "user must be manually added to 'lxd' group before using LXD"
    )


def test_is_initialized():
    """Return true if there is a valid disk device in the default profile."""
    mock_lxc = mock.Mock(spec=LXC)
    mock_lxc.profile_show.return_value = {
        "config": {},
        "description": "Default LXD profile",
        "devices": {
            "eth0": {
                "name": "eth0",
                "network": "lxdbr0",
                "type": "nic",
            },
            "root": {
                "path": "/",
                "pool": "default",
                "type": "disk",
            },
        },
        "name": "default",
    }

    initialized = is_initialized(lxc=mock_lxc, remote="test")

    assert initialized


def test_is_initialized_empty_profile():
    """Return false if the default profile is empty."""
    mock_lxc = mock.Mock(spec=LXC)
    mock_lxc.profile_show.return_value = {}

    initialized = is_initialized(lxc=mock_lxc, remote="test")

    assert not initialized


@pytest.mark.parametrize(
    "devices",
    [
        # no devices
        {},
        # path is not `/`
        {"root": {"path": "/foo", "pool": "default", "type": "disk"}},
        # type is not `disk`
        {"root": {"path": "/", "pool": "default", "type": "other"}},
    ],
)
def test_is_initialized_no_disk_device(devices):
    """Return false is there is not a valid disk device."""
    mock_lxc = mock.Mock(spec=LXC)
    mock_lxc.profile_show.return_value = {
        "config": {},
        "description": "Default LXD profile",
        "devices": devices,
        "name": "default",
    }

    initialized = is_initialized(lxc=mock_lxc, remote="test")

    assert not initialized


@pytest.mark.parametrize(
    ("which", "installed"), [("/path/to/lxd", True), (None, False)]
)
def test_is_installed(which, installed, monkeypatch):
    monkeypatch.setattr(shutil, "which", lambda x: which)

    assert is_installed() == installed


@pytest.mark.skipif(sys.platform != "linux", reason=f"unsupported on {sys.platform}")
def test_is_user_permitted(mock_os_access):
    mock_os_access.return_value = True

    assert is_user_permitted() is True


@pytest.mark.skipif(sys.platform != "linux", reason=f"unsupported on {sys.platform}")
def test_is_user_permitted_failure(mock_os_access):
    mock_os_access.return_value = False

    assert is_user_permitted() is False


def test_ensure_lxd_is_ready_not_installed(
    mock_lxd, mock_is_installed, mock_is_user_permitted, mock_is_initialized
):
    mock_is_installed.return_value = False

    with pytest.raises(LXDError) as exc_info:
        installer.ensure_lxd_is_ready(lxd=mock_lxd)

    assert exc_info.value == LXDError(
        brief="LXD is required, but not installed.",
        resolution=(
            "Visit https://documentation.ubuntu.com/lxd/en/latest/getting_started/ for "
            "instructions on installing and configuring LXD for your operating system."
        ),
    )


def test_ensure_lxd_is_ready_not_minimum_version(
    mock_lxd, mock_is_installed, mock_is_user_permitted, mock_is_initialized
):
    mock_lxd.is_supported_version.return_value = False
    mock_lxd.version.return_value = "3.12"

    with pytest.raises(LXDError) as exc_info:
        installer.ensure_lxd_is_ready(lxd=mock_lxd)

    assert exc_info.value == LXDError(
        brief="LXD '3.12' does not meet the minimum required version '4.0'.",
        resolution=(
            "Visit https://documentation.ubuntu.com/lxd/en/latest/getting_started/ for "
            "instructions on installing and configuring LXD for your operating system."
        ),
    )


def test_ensure_lxd_is_ready_not_permitted(
    mock_lxd, mock_is_installed, mock_is_user_permitted, mock_is_initialized
):
    mock_is_user_permitted.return_value = False

    with pytest.raises(LXDError) as exc_info:
        installer.ensure_lxd_is_ready(lxd=mock_lxd)

    assert exc_info.value == LXDError(
        brief="LXD requires additional permissions.",
        resolution="Ensure that the user is in the 'lxd' group.\n"
        "Visit https://documentation.ubuntu.com/lxd/en/latest/getting_started/ for "
        "instructions on installing and configuring LXD for your operating system.",
    )


def test_ensure_lxd_is_ready_not_initialized(
    mock_lxd, mock_is_installed, mock_is_user_permitted, mock_is_initialized
):
    mock_is_initialized.return_value = False

    with pytest.raises(LXDError) as exc_info:
        installer.ensure_lxd_is_ready(lxd=mock_lxd)

    assert exc_info.value == LXDError(
        brief="LXD has not been properly initialized.",
        details=(
            "The default LXD profile is empty or does not contain a disk device "
            "with a path of '/'."
        ),
        resolution="Execute 'lxd init --auto' to initialize LXD.\n"
        "Visit https://documentation.ubuntu.com/lxd/en/latest/getting_started/ for "
        "instructions on installing and configuring LXD for your operating system.",
    )


def test_ensure_lxd_is_ready_ok(
    mock_lxd, mock_is_installed, mock_is_user_permitted, mock_is_initialized
):
    installer.ensure_lxd_is_ready(lxd=mock_lxd)



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import subprocess
import textwrap

from craft_providers import errors


def test_provider_error():
    error = errors.ProviderError(brief="test brief")

    assert error == errors.ProviderError(brief="test brief")
    assert str(error) == "test brief"


def test_provider_error_with_details():
    error = errors.ProviderError(brief="test brief", details="test details")

    assert error == errors.ProviderError(brief="test brief", details="test details")
    assert str(error) == "test brief\ntest details"


def test_provider_error_with_resolution():
    error = errors.ProviderError(brief="test brief", resolution="test resolution")

    assert error == errors.ProviderError(
        brief="test brief", resolution="test resolution"
    )
    assert str(error) == "test brief\ntest resolution"


def test_provider_error_with_all():
    error = errors.ProviderError(
        brief="test brief", details="test details", resolution="test resolution"
    )

    assert error == errors.ProviderError(
        brief="test brief", details="test details", resolution="test resolution"
    )
    assert str(error) == "test brief\ntest details\ntest resolution"


def test_details_from_called_process_error():
    error = subprocess.CalledProcessError(
        -1, ["test-command", "flags", "quote$me"], "test stdout", "test stderr"
    )

    details = errors.details_from_called_process_error(error)

    assert details == textwrap.dedent(
        """\
            * Command that failed: "test-command flags 'quote$me'"
            * Command exit code: -1
            * Command output: 'test stdout'
            * Command standard error output: 'test stderr'"""
    )


def test_details_from_command_error():
    details = errors.details_from_command_error(
        returncode=-1,
        cmd=["test-command", "flags", "quote$me"],
    )

    assert details == textwrap.dedent(
        """\
            * Command that failed: "test-command flags 'quote$me'"
            * Command exit code: -1"""
    )


def test_details_from_command_error_with_output_strings():
    details = errors.details_from_command_error(
        returncode=-1,
        cmd=["test-command", "flags", "quote$me"],
        stdout="test stdout",
        stderr="test stderr",
    )

    assert details == textwrap.dedent(
        """\
            * Command that failed: "test-command flags 'quote$me'"
            * Command exit code: -1
            * Command output: 'test stdout'
            * Command standard error output: 'test stderr'"""
    )


def test_details_from_command_error_with_output_bytes():
    details = errors.details_from_command_error(
        returncode=-1,
        cmd=["test-command", "flags", "quote$me"],
        stdout=bytes.fromhex("00 FF"),
        stderr=bytes.fromhex("01 FE"),
    )

    assert details == textwrap.dedent(
        """\
            * Command that failed: "test-command flags 'quote$me'"
            * Command exit code: -1
            * Command output: b'\\x00\\xff'
            * Command standard error output: b'\\x01\\xfe'"""
    )



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Multipass Provider."""

import logging
import shutil
import subprocess
import sys
import time

from craft_providers.errors import details_from_called_process_error

from . import errors
from .multipass import Multipass

logger = logging.getLogger(__name__)


def install() -> str:
    """Install Multipass.

    :returns: Multipass version.

    :raises MultipassInstallationError: on error.
    """
    if sys.platform == "darwin":
        _install_darwin()
    elif sys.platform == "linux":
        _install_linux()
    elif sys.platform == "win32":
        raise errors.MultipassInstallationError(
            "automated installation not yet supported for Windows"
        )
    else:
        raise errors.MultipassInstallationError(
            f"unsupported platform {sys.platform!r}"
        )

    # TODO: Multipass needs time after being installed for `multipassd` to start.
    # Without a delay, errors could happen on launch, i.e.: "Remote "" is unknown or
    # unreachable." Current guidance is to sleep 20 seconds after install, but we
    # should have a more reliable and timely approach.
    # See: https://github.com/canonical/multipass/issues/1995
    time.sleep(20)

    multipass_version, _ = Multipass().wait_until_ready()
    return multipass_version


def _install_darwin() -> None:
    try:
        subprocess.run(["brew", "install", "multipass"], check=True)
        # wait for multipassd to start before changing the driver
        time.sleep(20)
        # this can be removed when multipass 1.12 is available on brew, because
        # qemu will be the new default
        subprocess.run(["multipass", "set", "local.driver=qemu"], check=True)
    except subprocess.CalledProcessError as error:
        raise errors.MultipassInstallationError(
            "error during brew installation",
            details=details_from_called_process_error(error),
        ) from error


def _install_linux() -> None:
    try:
        subprocess.run(["sudo", "snap", "install", "multipass"], check=True)
    except subprocess.CalledProcessError as error:
        raise errors.MultipassInstallationError(
            "error during snap installation",
            details=details_from_called_process_error(error),
        ) from error


def is_installed() -> bool:
    """Check if Multipass is installed (and found on PATH).

    :returns: True if multipass is installed.
    """
    return shutil.which("multipass") is not None



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import timedelta
from unittest.mock import call

import pytest
from craft_providers.bases import ubuntu
from craft_providers.errors import BaseConfigurationError
from craft_providers.lxd import LXDError, LXDProvider, LXDUnstableImageError


@pytest.fixture()
def mock_remote_image(mocker):
    _mock_remote_image = mocker.patch("craft_providers.lxd.remotes.RemoteImage")
    _mock_remote_image.image_name = "test-image-name"
    _mock_remote_image.remote_name = "test-remote-name"
    return _mock_remote_image


@pytest.fixture()
def mock_get_remote_image(mock_remote_image, mocker):
    _mock_get_remote_image = mocker.patch(
        "craft_providers.lxd.lxd_provider.get_remote_image",
        return_value=mock_remote_image,
    )
    return _mock_get_remote_image


@pytest.fixture()
def mock_buildd_base_configuration(mocker):
    mock_base_config = mocker.patch(
        "craft_providers.bases.ubuntu.BuilddBase", autospec=True
    )
    mock_base_config.alias = ubuntu.BuilddBaseAlias.JAMMY
    mock_base_config.compatibility_tag = "buildd-base-v2"
    return mock_base_config


@pytest.fixture()
def mock_lxc(mocker):
    return mocker.patch("craft_providers.lxd.LXC", autospec=True)


@pytest.fixture(autouse=True)
def mock_ensure_lxd_is_ready(mocker):
    return mocker.patch(
        "craft_providers.lxd.lxd_provider.ensure_lxd_is_ready", return_value=None
    )


@pytest.fixture()
def mock_install(mocker):
    return mocker.patch("craft_providers.lxd.lxd_provider.install")


@pytest.fixture(autouse=True)
def mock_is_installed(mocker):
    return mocker.patch(
        "craft_providers.lxd.lxd_provider.is_installed", return_value=True
    )


@pytest.fixture()
def mock_launch(mocker):
    return mocker.patch("craft_providers.lxd.lxd_provider.launch", autospec=True)


def test_ensure_provider_is_available_installed(
    mock_is_installed, mock_install, mock_ensure_lxd_is_ready
):
    """Verify LXD is installed if it is not already installed."""
    mock_is_installed.return_value = True
    provider = LXDProvider()

    provider.ensure_provider_is_available()

    mock_install.assert_not_called()
    mock_ensure_lxd_is_ready.assert_called_once()


def test_ensure_provider_is_available_not_installed(
    mock_is_installed, mock_install, mock_ensure_lxd_is_ready
):
    """Verify LXD is not re-installed if it is already installed."""
    mock_is_installed.return_value = False
    provider = LXDProvider()

    provider.ensure_provider_is_available()

    mock_install.assert_called_once()
    mock_ensure_lxd_is_ready.assert_called_once()


@pytest.mark.parametrize("is_installed", [True, False])
def test_is_provider_installed(is_installed, mock_is_installed):
    """Verify `is_provider_installed` checks if LXD is actually installed."""
    mock_is_installed.return_value = is_installed
    provider = LXDProvider()

    assert provider.is_provider_installed() == is_installed


def test_create_environment(mocker):
    mock_lxd_instance = mocker.patch("craft_providers.lxd.lxd_provider.LXDInstance")

    provider = LXDProvider()
    provider.create_environment(instance_name="test-name")

    mock_lxd_instance.assert_called_once_with(
        name="test-name", project="default", remote="local"
    )


@pytest.mark.parametrize(
    ("allow_unstable", "is_stable"),
    [
        # all permutations are valid except allow_unstable=False and is_stable=False
        (True, True),
        (True, False),
        (False, True),
    ],
)
def test_launched_environment(
    allow_unstable,
    is_stable,
    mock_buildd_base_configuration,
    mock_get_remote_image,
    mock_remote_image,
    mock_launch,
    mock_lxc,
    tmp_path,
):
    mock_remote_image.is_stable = is_stable
    provider = LXDProvider(lxc=mock_lxc)

    # set the expected expiration time
    expiration = timedelta(days=90) if is_stable else timedelta(days=14)

    with provider.launched_environment(
        project_name="test-project",
        project_path=tmp_path,
        base_configuration=mock_buildd_base_configuration,
        instance_name="test-instance-name",
        allow_unstable=allow_unstable,
    ) as instance:
        assert instance is not None
        mock_get_remote_image.assert_called_once_with(mock_buildd_base_configuration)
        mock_remote_image.add_remote.assert_called_once_with(lxc=mock_lxc)
        assert mock_launch.mock_calls == [
            call(
                name="test-instance-name",
                base_configuration=mock_buildd_base_configuration,
                image_name="test-image-name",
                image_remote="test-remote-name",
                auto_clean=True,
                auto_create_project=True,
                map_user_uid=True,
                uid=tmp_path.stat().st_uid,
                use_base_instance=True,
                project="default",
                remote="local",
                expiration=expiration,
            ),
        ]

        mock_launch.reset_mock()

    assert mock_launch.mock_calls == [
        call().unmount_all(),
        call().stop(),
    ]


def test_launched_environment_launch_base_configuration_error(
    mock_buildd_base_configuration,
    mock_get_remote_image,
    mock_remote_image,
    mock_launch,
    tmp_path,
):
    error = BaseConfigurationError(brief="fail")
    mock_launch.side_effect = error
    provider = LXDProvider()

    with pytest.raises(LXDError, match="fail") as raised, provider.launched_environment(
        project_name="test-project",
        project_path=tmp_path,
        base_configuration=mock_buildd_base_configuration,
        instance_name="test-instance-name",
    ):
        pass

    assert raised.value.__cause__ is error


def test_launched_environment_unstable_error(
    mock_buildd_base_configuration,
    mock_get_remote_image,
    mock_remote_image,
    mock_launch,
    tmp_path,
):
    """Raise an Exception when an unstable image is used with opting in."""
    mock_remote_image.is_stable = False
    provider = LXDProvider()

    with pytest.raises(LXDUnstableImageError) as raised, provider.launched_environment(
        project_name="test-project",
        project_path=tmp_path,
        base_configuration=mock_buildd_base_configuration,
        instance_name="test-instance-name",
    ):
        pass

    assert raised.value.brief == (
        "Cannot launch an unstable image 'test-image-name' from remote "
        "'test-remote-name'"
    )


def test_lxd_name():
    """Verify LXDProvider's name."""
    provider = LXDProvider()

    assert provider.name == "LXD"


def test_lxd_install_recommendation():
    """Verify LXDProvider's install recommendation."""
    provider = LXDProvider()

    assert (
        provider.install_recommendation
        == "Visit https://ubuntu.com/lxd/install for instructions to install LXD."
    )



# Copyright 2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Craft Providers versioning tests."""
import re
import subprocess

import craft_providers
import pytest


def _repo_has_version_tag() -> bool:
    """Returns True if the repo has a git tag usable for versioning."""
    git_describe_command = [
        "git",
        "describe",
        "--always",
        "--long",
        "--match",
        "[0-9]*.[0-9]*.[0-9]*",
        "--exclude",
        "*[^0-9.]*",
    ]
    output = subprocess.run(
        git_describe_command, check=True, capture_output=True, text=True
    ).stdout.rstrip("\n")

    # match on 'X.Y.Z-<commits since tag>-g<hash>'
    return bool(re.fullmatch(r"\d+\.\d+\.\d+-\d+-g[0-9a-f]+", output))


@pytest.mark.skipif(
    _repo_has_version_tag(), reason="Skipping because project was versioned from a tag."
)
def test_version_without_tags():
    """Validate version format when no valid tag are present."""
    version = craft_providers.__version__

    # match on '0.0.post<total commits>+g<hash>'
    #       or '0.0.post<total commits>+g<hash>.d<%Y%m%d>'
    assert re.fullmatch(r"\d+\.\d+\.post\d+\+g[0-9a-f]+(\.d[0-9]*)?", version)


@pytest.mark.skipif(
    not _repo_has_version_tag(),
    reason="Skipping because project was not versioned from a tag.",
)
def test_version_with_tags():
    """Version should be properly formatted when a valid tag exists."""
    version = craft_providers.__version__

    # match on 'X.Y.Z'
    #       or 'X.Y.Z.d<%Y%m%d>'
    #       or 'X.Y.Z.post<commits since tag>+g<hash>'
    #       or 'X.Y.Z.post<commits since tag>+g<hash>.d<%Y%m%d>'
    assert re.fullmatch(r"\d+\.\d+\.\d+(\.post\d+\+g[0-9a-f]+)?(\.d[0-9]*)?", version)



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Craft Providers base package."""

from .base import Base
from .errors import ProviderError
from .executor import Executor
from .provider import Provider

try:
    from ._version import __version__
except ImportError:  # pragma: no cover
    from importlib.metadata import version, PackageNotFoundError

    try:
        __version__ = version("craft_providers")
    except PackageNotFoundError:
        __version__ = "dev"

__all__ = [
    "__version__",
    "Base",
    "Executor",
    "ProviderError",
    "Provider",
]



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import io
import pathlib
import subprocess

import pytest
from craft_providers import multipass
from craft_providers.bases import ubuntu
from craft_providers.errors import BaseCompatibilityError

from . import conftest

pytestmark = [pytest.mark.xdist_group("multipass_launch_tests")]


@pytest.fixture()
def core22_instance(instance_name):
    """Yields a minimally setup core22 instance.

    The yielded instance will be launched, started, and marked as setup, even though
    most of the setup is skipped to speed up test execution.

    Delete instance on fixture teardown.
    """
    with conftest.tmp_instance(
        instance_name=instance_name,
        image_name="snapcraft:core22",
    ) as tmp_instance:
        instance = multipass.MultipassInstance(name=tmp_instance)

        # mark instance as setup in the config file
        instance.push_file_io(
            destination=ubuntu.BuilddBase._instance_config_path,
            content=io.BytesIO(
                f"compatibility_tag: {ubuntu.BuilddBase.compatibility_tag}"
                "\nsetup: true\n".encode()
            ),
            file_mode="0644",
        )

        yield instance

        if instance.exists():
            instance.delete()


def test_launch(instance_name):
    """Launch an instance and run a command inside the instance."""
    base_configuration = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    instance = multipass.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="snapcraft:core22",
        disk_gb=16,
    )

    try:
        assert isinstance(instance, multipass.MultipassInstance)
        assert instance.exists() is True
        assert instance.is_running() is True

        proc = instance.execute_run(["echo", "hi"], check=True, stdout=subprocess.PIPE)

        assert proc.stdout == b"hi\n"
    finally:
        instance.delete()


def test_launch_existing_instance(core22_instance):
    base_configuration = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    instance = multipass.launch(
        name=core22_instance.name,
        base_configuration=base_configuration,
        image_name="snapcraft:core22",
        disk_gb=16,
    )

    assert isinstance(instance, multipass.MultipassInstance)
    assert instance.exists() is True
    assert instance.is_running() is True

    proc = instance.execute_run(["echo", "hi"], check=True, stdout=subprocess.PIPE)

    assert proc.stdout == b"hi\n"


def test_launch_os_incompatible_instance(core22_instance):
    base_configuration = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    core22_instance.push_file_io(
        destination=pathlib.PurePosixPath("/etc/os-release"),
        content=io.BytesIO(b"NAME=Fedora\nVERSION_ID=32\n"),
        file_mode="0644",
    )

    # Should raise compatibility error with auto_clean=False.
    with pytest.raises(BaseCompatibilityError) as exc_info:
        multipass.launch(
            name=core22_instance.name,
            base_configuration=base_configuration,
            image_name="snapcraft:core22",
        )

    assert (
        exc_info.value.brief
        == "Incompatible base detected: Expected OS 'Ubuntu', found 'Fedora'."
    )

    # Retry with auto_clean=True.
    multipass.launch(
        name=core22_instance.name,
        base_configuration=base_configuration,
        image_name="snapcraft:core22",
        disk_gb=16,
        auto_clean=True,
    )

    assert core22_instance.exists() is True
    assert core22_instance.is_running() is True


def test_launch_instance_config_incompatible_instance(core22_instance):
    base_configuration = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    core22_instance.push_file_io(
        destination=base_configuration._instance_config_path,
        content=io.BytesIO(b"compatibility_tag: invalid\nsetup: true\n"),
        file_mode="0644",
    )

    # Should raise compatibility error with auto_clean=False.
    with pytest.raises(BaseCompatibilityError) as exc_info:
        multipass.launch(
            name=core22_instance.name,
            base_configuration=base_configuration,
            image_name="snapcraft:core22",
        )

    assert exc_info.value.brief == (
        "Incompatible base detected:"
        " Expected image compatibility tag 'buildd-base-v7', found 'invalid'."
    )

    # Retry with auto_clean=True.
    multipass.launch(
        name=core22_instance.name,
        base_configuration=base_configuration,
        image_name="snapcraft:core22",
        disk_gb=16,
        auto_clean=True,
    )

    assert core22_instance.exists() is True
    assert core22_instance.is_running() is True


def test_launch_instance_not_setup_without_auto_clean(core22_instance):
    """Raise an error if an existing instance is not setup and auto_clean is False."""
    base_configuration = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    core22_instance.push_file_io(
        destination=base_configuration._instance_config_path,
        content=io.BytesIO(b"compatibility_tag: buildd-base-v7\nsetup: false\n"),
        file_mode="0644",
    )

    # will raise a compatibility error
    with pytest.raises(BaseCompatibilityError) as exc_info:
        multipass.launch(
            name=core22_instance.name,
            base_configuration=base_configuration,
            image_name="snapcraft:core22",
            auto_clean=False,
        )

    assert exc_info.value == BaseCompatibilityError("instance is marked as not setup")


def test_launch_instance_not_setup_with_auto_clean(core22_instance):
    """Clean the instance if it is not setup and auto_clean is True."""
    base_configuration = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    core22_instance.push_file_io(
        destination=base_configuration._instance_config_path,
        content=io.BytesIO(b"compatibility_tag: buildd-base-v7\nsetup: false\n"),
        file_mode="0644",
    )

    # when auto_clean is true, the instance will be deleted and recreated
    multipass.launch(
        name=core22_instance.name,
        base_configuration=base_configuration,
        image_name="snapcraft:core22",
        disk_gb=16,
        auto_clean=True,
    )

    assert core22_instance.exists() is True
    assert core22_instance.is_running() is True



#
# Copyright 2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Almalinux image(s)."""
import enum
import logging
import pathlib
import subprocess
from typing import Dict, List, Optional

from craft_providers.actions.snap_installer import Snap
from craft_providers.base import Base
from craft_providers.errors import (
    BaseCompatibilityError,
    BaseConfigurationError,
    details_from_called_process_error,
)
from craft_providers.executor import Executor

logger = logging.getLogger(__name__)


class AlmaLinuxBaseAlias(enum.Enum):
    """Mappings for supported AlmaLinux images."""

    NINE = "9"


class AlmaLinuxBase(Base):
    """Support for AlmaLinux images.

    :cvar compatibility_tag: Tag/Version for variant of build configuration and
        setup.  Any change to this version would indicate that prior [versioned]
        instances are incompatible and must be cleaned.  As such, any new value
        should be unique to old values (e.g. incrementing).  It is suggested to
        extend this tag, not overwrite it, e.g.: compatibility_tag =
        f"{appname}-{BuildBase.compatibility_tag}.{apprevision}" to ensure base
        compatibility levels are maintained.
    :cvar instance_config_path: Path to persistent environment configuration
        used for compatibility checks (or other data).  Set to
        /etc/craft-instance.conf, but may be overridden for application-specific
        reasons.
    :cvar instance_config_class: Class defining instance configuration.  May be
        overridden with an application-specific subclass of InstanceConfiguration
        to enable application-specific extensions.

    :param alias: Base alias / version.
    :param environment: Environment to set in /etc/environment.
    :param hostname: Hostname to configure.
    :param snaps: Optional list of snaps to install on the base image.
    :param packages: Optional list of system packages to install on the base image.
    :param use_default_packages: Optional bool to enable/disable default packages.
    :param cache_path: (Optional) Path to the shared cache directory. If this is
        provided, shared cache directories will be mounted as appropriate.
    """

    compatibility_tag: str = f"almalinux-{Base.compatibility_tag}"

    def __init__(
        self,
        *,
        alias: AlmaLinuxBaseAlias,
        compatibility_tag: Optional[str] = None,
        environment: Optional[Dict[str, Optional[str]]] = None,
        hostname: str = "craft-almalinux-instance",
        snaps: Optional[List[Snap]] = None,
        packages: Optional[List[str]] = None,
        use_default_packages: bool = True,
        cache_path: Optional[pathlib.Path] = None,
    ) -> None:
        self._cache_path = cache_path

        # ignore enum subclass (see https://github.com/microsoft/pyright/issues/6750)
        self.alias: AlmaLinuxBaseAlias = alias  # pyright: ignore

        if environment is None:
            self._environment = self.default_command_environment()
        else:
            self._environment = environment

        if compatibility_tag:
            self.compatibility_tag = compatibility_tag

        self._set_hostname(hostname)

        self._packages: Optional[List[str]] = []
        if use_default_packages:
            self._packages.extend(
                [
                    "autoconf",
                    "automake",
                    "gcc",
                    "gcc-c++",
                    "git",
                    "make",
                    "patch",
                    "python3",
                    "python3-devel",
                    "python3-pip",
                    "python3-pip-wheel",
                    "python3-setuptools",
                ]
            )

        if packages:
            self._packages.extend(packages)

        self._snaps = snaps

    def _ensure_os_compatible(self, executor: Executor) -> None:
        """Ensure OS is compatible with Base.

        :raises BaseCompatibilityError: if instance is incompatible.
        :raises BaseConfigurationError: on other unexpected error.
        """
        os_release = self._get_os_release(executor=executor)

        os_id = os_release.get("ID")
        if os_id != "almalinux":
            raise BaseCompatibilityError(
                reason=f"Expected OS 'almalinux', found {os_id!r}"
            )

        compat_version_id = self.alias.value
        version_id = os_release.get("VERSION_ID", "")
        version_id = version_id.split(".")[0]

        if version_id != compat_version_id:
            raise BaseCompatibilityError(
                reason=(
                    f"Expected OS version {compat_version_id!r},"
                    f" found {version_id!r}"
                )
            )

    def _enable_dnf_extra_repos(self, executor: Executor) -> None:
        """Configure AlmaLinux special extra repos.

        Enable "epel-release" repo for snapd.
        """
        try:
            command = ["dnf", "install", "-y", "epel-release"]
            self._execute_run(
                command,
                executor=executor,
                verify_network=True,
                timeout=self._timeout_complex,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to enable extra repos.",
                details=details_from_called_process_error(error),
            ) from error

    def _pre_setup_packages(self, executor: Executor) -> None:
        """Configure dnf package manager."""
        self._enable_dnf_extra_repos(executor=executor)

    def _setup_packages(self, executor: Executor) -> None:
        """Install needed packages using dnf."""
        # update system
        try:
            self._execute_run(
                ["dnf", "update", "-y"],
                executor=executor,
                verify_network=True,
                timeout=self._timeout_unpredictable,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to update system using dnf.",
                details=details_from_called_process_error(error),
            ) from error

        # install required packages and user-defined packages
        if not self._packages:
            return
        try:
            command = ["dnf", "install", "-y", *self._packages]
            self._execute_run(
                command,
                executor=executor,
                verify_network=True,
                timeout=self._timeout_unpredictable,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to install packages.",
                details=details_from_called_process_error(error),
            ) from error

    def _setup_snapd(self, executor: Executor) -> None:
        """Set up snapd using dnf."""
        try:
            self._execute_run(
                ["dnf", "install", "-y", "snapd"],
                executor=executor,
                verify_network=True,
                timeout=self._timeout_complex,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to setup snapd.",
                details=details_from_called_process_error(error),
            ) from error

    def _clean_up(self, executor: Executor) -> None:
        """Clean up unused packages and cached package files."""
        self._execute_run(
            ["dnf", "autoremove", "-y"],
            executor=executor,
            timeout=self._timeout_complex,
        )
        self._execute_run(
            ["dnf", "clean", "packages", "-y"],
            executor=executor,
            timeout=self._timeout_complex,
        )



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import shutil
import sys
from unittest import mock

import pytest
from craft_providers import multipass

# Shortcut any calls to time.sleep with pytest-time's instant_sleep
pytestmark = pytest.mark.usefixtures("instant_sleep")


@pytest.fixture()
def mock_details_from_process_error():
    details = "<details>"
    with mock.patch(
        "craft_providers.multipass.installer.details_from_called_process_error",
        return_value=details,
    ) as mock_details:
        yield mock_details


def test_install_darwin(fake_process, monkeypatch, mock_instant_sleep):
    monkeypatch.setattr(sys, "platform", "darwin")

    fake_process.register_subprocess(["brew", "install", "multipass"])
    fake_process.register_subprocess(["multipass", "set", "local.driver=qemu"])
    fake_process.register_subprocess(
        ["multipass", "version"], stdout="multipass 1.4.2\nmultipassd 1.4.2\n"
    )

    multipass.install()

    assert list(fake_process.calls) == [
        ["brew", "install", "multipass"],
        ["multipass", "set", "local.driver=qemu"],
        ["multipass", "version"],
    ]
    mock_instant_sleep.sleep.assert_called_with(20)


def test_install_darwin_error(
    fake_process, mock_details_from_process_error, monkeypatch
):
    monkeypatch.setattr(sys, "platform", "darwin")
    fake_process.register_subprocess(
        ["brew", "install", "multipass"],
        returncode=1,
    )

    with pytest.raises(multipass.MultipassInstallationError) as exc_info:
        multipass.install()

    assert exc_info.value == multipass.MultipassInstallationError(
        reason="error during brew installation",
        details=mock_details_from_process_error.return_value,
    )


def test_install_linux(fake_process, monkeypatch):
    monkeypatch.setattr(sys, "platform", "linux")

    fake_process.register_subprocess(["sudo", "snap", "install", "multipass"])
    fake_process.register_subprocess(
        ["multipass", "version"], stdout="multipass 1.4.2\nmultipassd 1.4.2\n"
    )

    multipass.install()

    assert list(fake_process.calls) == [
        ["sudo", "snap", "install", "multipass"],
        ["multipass", "version"],
    ]


def test_install_linux_error(
    fake_process, mock_details_from_process_error, monkeypatch
):
    monkeypatch.setattr(sys, "platform", "linux")
    fake_process.register_subprocess(
        ["sudo", "snap", "install", "multipass"],
        returncode=1,
    )

    with pytest.raises(multipass.MultipassInstallationError) as exc_info:
        multipass.install()

    assert exc_info.value == multipass.MultipassInstallationError(
        reason="error during snap installation",
        details=mock_details_from_process_error.return_value,
    )


def test_install_windows_error(monkeypatch):
    monkeypatch.setattr(sys, "platform", "win32")

    with pytest.raises(multipass.MultipassInstallationError) as exc_info:
        multipass.install()

    assert exc_info.value == multipass.MultipassInstallationError(
        reason="automated installation not yet supported for Windows",
    )


@pytest.mark.parametrize(
    ("which", "is_installed"), [("/path/to/multipass", True), (None, False)]
)
def test_is_installed(which, is_installed, monkeypatch):
    monkeypatch.setattr(shutil, "which", lambda x: which)

    assert multipass.is_installed() == is_installed



#
# Copyright 2021-2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#


from craft_providers.util import snap_cmd


def test_ack(tmp_path):
    command = snap_cmd.formulate_ack_command(snap_assert_path=tmp_path)
    assert command == ["snap", "ack", tmp_path.as_posix()]


def test_known(tmp_path):
    command = snap_cmd.formulate_known_command(query=["test1", "test2"])
    assert command == ["snap", "known", "test1", "test2"]


def test_local_install_strict(tmp_path):
    assert snap_cmd.formulate_local_install_command(
        classic=False, dangerous=False, snap_path=tmp_path
    ) == ["snap", "install", tmp_path.as_posix()]


def test_local_install_classic(tmp_path):
    assert snap_cmd.formulate_local_install_command(
        classic=True, dangerous=False, snap_path=tmp_path
    ) == ["snap", "install", tmp_path.as_posix(), "--classic"]


def test_local_install_dangerous(tmp_path):
    assert snap_cmd.formulate_local_install_command(
        classic=False, dangerous=True, snap_path=tmp_path
    ) == ["snap", "install", tmp_path.as_posix(), "--dangerous"]


def test_local_install_all_opts(tmp_path):
    assert snap_cmd.formulate_local_install_command(
        classic=True, dangerous=True, snap_path=tmp_path
    ) == ["snap", "install", tmp_path.as_posix(), "--classic", "--dangerous"]


def test_pack():
    command = snap_cmd.formulate_pack_command(
        snap_name="testsnap", output_file_path="testpath"
    )
    assert command == ["snap", "pack", "/snap/testsnap/current/", "--filename=testpath"]


def test_remote_install_strict():
    snap_name, channel = "testsnap", "edge"
    cmd = snap_cmd.formulate_remote_install_command(snap_name, channel, classic=False)
    assert cmd == ["snap", "install", snap_name, "--channel", channel]


def test_remote_install_classic():
    snap_name, channel = "testsnap", "edge"
    cmd = snap_cmd.formulate_remote_install_command(snap_name, channel, classic=True)
    assert cmd == ["snap", "install", snap_name, "--channel", channel, "--classic"]


def test_refresh():
    snap_name, channel = "testsnap", "edge"
    cmd = snap_cmd.formulate_refresh_command(snap_name, channel)
    assert cmd == ["snap", "refresh", snap_name, "--channel", channel]


def test_remove():
    snap_name = "testsnap"
    cmd = snap_cmd.formulate_remove_command(snap_name)
    assert cmd == ["snap", "remove", snap_name]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from craft_providers import ProviderError
from craft_providers.lxd import LXDError, LXDProvider
from craft_providers.multipass import MultipassError, MultipassProvider

known_provider_classes = [LXDProvider, MultipassProvider]


@pytest.fixture(params=known_provider_classes)
def provider_class(request):
    return request.param


@pytest.fixture(autouse=True)
def mock_lxd_delete(mocker):
    return mocker.patch("craft_providers.lxd.LXDInstance.delete")


@pytest.fixture(autouse=True)
def mock_lxd_exists(mocker):
    return mocker.patch("craft_providers.lxd.LXDInstance.exists", return_value=True)


@pytest.fixture(autouse=True)
def mock_lxd_is_installed(mocker):
    return mocker.patch(
        "craft_providers.lxd.lxd_provider.is_installed", return_value=True
    )


@pytest.fixture(autouse=True)
def mock_multipass_delete(mocker):
    return mocker.patch("craft_providers.multipass.MultipassInstance.delete")


@pytest.fixture(autouse=True)
def mock_multipass_exists(mocker):
    return mocker.patch(
        "craft_providers.multipass.MultipassInstance.exists", return_value=True
    )


@pytest.fixture(autouse=True)
def mock_multipass_is_installed(mocker):
    return mocker.patch(
        "craft_providers.multipass.multipass_provider.is_installed", return_value=True
    )


def test_clean_project_environment_exists(
    mock_lxd_delete, mock_multipass_delete, provider_class
):
    """Assert instance is deleted if it exists."""
    provider = provider_class()
    provider.clean_project_environments(instance_name="test-name")

    if isinstance(provider, LXDProvider):
        mock_lxd_delete.assert_called_once()
    else:
        mock_multipass_delete.assert_called_once()


def test_clean_project_environment_does_not_exist(
    mock_lxd_exists,
    mock_lxd_delete,
    mock_multipass_exists,
    mock_multipass_delete,
    provider_class,
):
    """Assert instance is not deleted if it does not exist."""
    mock_lxd_exists.return_value = False
    mock_multipass_exists.return_value = False

    provider = provider_class()
    provider.clean_project_environments(instance_name="test-name")

    mock_lxd_delete.assert_not_called()
    mock_multipass_delete.assert_not_called()


def test_clean_project_environment_not_installed(
    mock_lxd_delete,
    mock_lxd_is_installed,
    mock_multipass_delete,
    mock_multipass_is_installed,
    provider_class,
):
    """Assert instance is not deleted if the provider is not installed."""
    mock_lxd_is_installed.return_value = False
    mock_multipass_is_installed.return_value = False

    provider = provider_class()
    provider.clean_project_environments(instance_name="test-name")

    mock_lxd_delete.assert_not_called()
    mock_multipass_delete.assert_not_called()


def test_clean_project_environment_exists_error(
    mock_lxd_exists, mock_multipass_exists, provider_class
):
    """Assert error on `exists` call is caught."""
    mock_lxd_exists.side_effect = LXDError("fail")
    mock_multipass_exists.side_effect = MultipassError("fail")
    provider = provider_class()

    with pytest.raises(ProviderError) as raised:
        provider.clean_project_environments(instance_name="test-name")

    assert str(raised.value) == "fail"


def test_clean_project_environment_delete_error(
    mock_lxd_delete, mock_multipass_delete, provider_class
):
    """Assert error on `delete` call is caught."""
    mock_lxd_delete.side_effect = LXDError("fail")
    mock_multipass_delete.side_effect = MultipassError("fail")
    provider = provider_class()

    with pytest.raises(ProviderError) as raised:
        provider.clean_project_environments(instance_name="test-name")

    assert str(raised.value) == "fail"



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Fixtures for integration tests."""

import contextlib
import os
import pathlib
import random
import re
import shutil
import string
import subprocess
import sys
import tempfile
from typing import Optional

import pytest
from craft_providers import lxd, multipass
from craft_providers.actions.snap_installer import get_host_snap_info
from craft_providers.bases import ubuntu


def pytest_runtest_setup(item: pytest.Item):
    """Configuration for tests."""
    with_sudo = item.get_closest_marker("with_sudo")
    if (
        with_sudo
        and not os.getenv("CI")
        and not os.getenv("CRAFT_PROVIDERS_TESTS_ENABLE_SUDO")
    ):
        pytest.skip("Not running in CI and CRAFT_PROVIDERS_TESTS_ENABLE_SUDO not set.")


def generate_instance_name():
    """Generate a random instance name."""
    return "itest-" + "".join(random.choices(string.ascii_uppercase, k=8))


def snap_exists(snap_name: str) -> bool:
    """Returns true if a snap exists."""
    return os.path.exists(f"/snap/{snap_name}/current")


def is_installed_dangerously(snap_name: str) -> bool:
    """Returns true if a snap is installed dangerously."""
    return get_host_snap_info(snap_name)["revision"].startswith("x")


@pytest.fixture()
def home_tmp_path():
    """Provide a temporary directory located in user's home directory.

    Multipass doesn't have access to /tmp, this fixture provides an equivalent
    to tmp_path for tests that require it.
    """
    with tempfile.TemporaryDirectory(
        suffix=".tmp-pytest", dir=pathlib.Path.home()
    ) as temp_dir:
        yield pathlib.Path(temp_dir)


@pytest.fixture()
def instance_name():
    """Provide a random name for an instance to launch."""
    return generate_instance_name()


@pytest.fixture(scope="module")
def reusable_instance_name():
    """Provide a random name for an instance to launch with scope=module."""
    return generate_instance_name()


@pytest.fixture(scope="session")
def installed_lxd():
    """Ensure lxd is installed, or skip the test if we cannot.

    If the environment has CRAFT_PROVIDERS_TESTS_ENABLE_LXD_INSTALL=1,
    force the installation of LXD if uninstalled.
    """
    if sys.platform != "linux":
        pytest.skip(f"lxd not supported on {sys.platform}")

    if lxd.is_installed():
        return

    if os.environ.get("CRAFT_PROVIDERS_TESTS_ENABLE_LXD_INSTALL") == "1":
        lxd.install()
    else:
        pytest.skip("lxd not installed, skipped")


@pytest.fixture()
def uninstalled_lxd():
    """Uninstall Lxd prior to test, if environment allows it.

    For consistency with installed_lxd fixture, LXD must be installed prior to
    this, signaling that LXD will run OK.

    Environment may enable this fixture with:
    CRAFT_PROVIDERS_TESTS_ENABLE_LXD_UNINSTALL=1
    """
    if sys.platform != "linux":
        pytest.skip(f"lxd not supported on {sys.platform}")

    if not lxd.is_installed():
        pytest.skip("lxd not installed, skipped")

    if os.environ.get("CRAFT_PROVIDERS_TESTS_ENABLE_LXD_UNINSTALL") != "1":
        pytest.skip("not configured to uninstall lxd, skipped")

    if sys.platform == "linux":
        subprocess.run(["sudo", "snap", "remove", "lxd", "--purge"], check=True)

    yield

    # Ensure it is installed after test.
    if not lxd.is_installed():
        lxd.install()


@pytest.fixture(scope="session")
def installed_multipass():
    """Ensure multipass is installed, or skip the test if we cannot.

    If the environment has CRAFT_PROVIDERS_TESTS_ENABLE_MULTIPASS_INSTALL=1,
    force the installation of Multipass if uninstalled.
    """
    if multipass.is_installed():
        return

    if os.environ.get("CRAFT_PROVIDERS_TESTS_ENABLE_MULTIPASS_INSTALL") == "1":
        multipass.install()
    else:
        pytest.skip("multipass not installed, skipped")


@pytest.fixture()
def uninstalled_multipass():
    """Uninstall Multipass prior to test, if environment allows it.

    For consistency with installed_multipass fixture, Multipass must be
    installed prior to this, signaling that Multipass will run OK.

    Environment may enable this fixture with:
    CRAFT_PROVIDERS_TESTS_ENABLE_MULTIPASS_UNINSTALL=1
    """
    if not multipass.is_installed():
        pytest.skip("multipass not installed, skipped")

    if os.environ.get("CRAFT_PROVIDERS_TESTS_ENABLE_MULTIPASS_UNINSTALL") != "1":
        pytest.skip("not configured to uninstall multipass, skipped")

    if sys.platform == "linux":
        subprocess.run(["sudo", "snap", "remove", "multipass", "--purge"], check=True)
    elif sys.platform == "darwin":
        subprocess.run(["brew", "uninstall", "multipass"], check=True)
    else:
        pytest.skip("platform not supported to uninstall multipass, skipped")

    yield

    # Ensure it is installed after test.
    if not multipass.is_installed():
        multipass.install()


@pytest.fixture()
def core22_lxd_instance(installed_lxd, instance_name):
    """Fully configured buildd-based core22 LXD instance."""
    base_configuration = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        ephemeral=True,
    )

    yield instance

    if instance.exists():
        instance.delete()


@pytest.fixture(scope="session")
def installed_snap():
    """Fixture to provide contextmanager to install a specified snap.

    If a snap is not installed, it would be installed automatically with:
    CRAFT_PROVIDERS_TESTS_ENABLE_SNAP_INSTALL=1
    """

    if shutil.which("snap") is None or sys.platform != "linux":
        pytest.skip("requires linux and snapd")

    @contextlib.contextmanager
    def _installed_snap(snap_name, *, try_path: Optional[pathlib.Path] = None):
        """Ensure snap is installed or skip test."""
        # do nothing if already installed and not dangerous
        if snap_exists(snap_name) and not is_installed_dangerously(snap_name):
            yield
        else:
            # Install it, if enabled to do so by environment.
            if os.environ.get("CRAFT_PROVIDERS_TESTS_ENABLE_SNAP_INSTALL") != "1":
                pytest.skip(f"{snap_name!r} snap not installed, skipped")

            if try_path:
                subprocess.run(["sudo", "snap", "try", str(try_path)], check=True)
            # if snap is already installed dangerously, use 'snap refresh'
            elif snap_exists(snap_name) and is_installed_dangerously(snap_name):
                subprocess.run(
                    ["sudo", "snap", "refresh", "--amend", snap_name], check=True
                )
            # else use 'snap install'
            else:
                subprocess.run(["sudo", "snap", "install", snap_name], check=True)
            yield
            subprocess.run(["sudo", "snap", "remove", snap_name], check=True)

    return _installed_snap


@pytest.fixture()
def dangerously_installed_snap(tmpdir):
    """Fixture to provide contextmanager for a dangerously installed snap.

    If the snap is not installed, it would be installed automatically with:
    CRAFT_PROVIDERS_TESTS_ENABLE_SNAP_INSTALL=1
    """

    @contextlib.contextmanager
    def _dangerously_installed_snap(snap_name):
        """Ensure snap is installed or skip test."""
        if shutil.which("snap") is None or sys.platform != "linux":
            pytest.skip("requires linux and snapd")

        # do nothing if the snap is already installed dangerously
        if snap_exists(snap_name) and is_installed_dangerously(snap_name):
            yield
        else:
            if os.environ.get("CRAFT_PROVIDERS_TESTS_ENABLE_SNAP_INSTALL") != "1":
                pytest.skip(f"{snap_name!r} snap not installed, skipped")

            # download the snap
            output = subprocess.run(
                ["snap", "download", "--target-directory", tmpdir, snap_name],
                check=True,
                capture_output=True,
            )

            # collect the file name
            match = re.search(f"{snap_name}_\\d+.snap", str(output))
            if not match:
                raise Exception(
                    "could not parse snap file name from output of "
                    f"'snap download {snap_name}' (output = {output!r})"
                )
            snap_file_path = pathlib.Path(tmpdir / match.group())

            # the `--dangerous` flag will force the snap to be installed dangerously,
            # even if the assertions exist in snapd
            subprocess.run(
                ["sudo", "snap", "install", snap_file_path, "--dangerous"], check=True
            )
            yield
            subprocess.run(["sudo", "snap", "remove", snap_name], check=True)

    return _dangerously_installed_snap


@pytest.fixture(scope="session")
def empty_test_snap(installed_snap):
    """Fixture to provide an empty local-only snap for test purposes.

    Requires:
    CRAFT_PROVIDERS_TESTS_ENABLE_SNAP_INSTALL=1
    """
    snap_name = "craft-integration-test-snap"

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir)
        tmp_path.chmod(0o755)

        meta_dir = tmp_path / "meta"
        meta_dir.mkdir()
        snap_yaml = meta_dir / "snap.yaml"
        snap_yaml.write_text(
            f"name: {snap_name}\nversion: 1.0\ntype: base\nsummary: test snap\n"
        )

        with installed_snap(snap_name, try_path=tmp_path):
            yield snap_name



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import io
import subprocess

import pytest
from craft_providers.multipass import Multipass

from . import conftest


@pytest.fixture()
def instance(instance_name):
    with conftest.tmp_instance(
        instance_name=instance_name,
    ) as tmp_instance:
        yield tmp_instance


@pytest.fixture()
def multipass():
    return Multipass()


def test_delete(instance, multipass):
    multipass.delete(instance_name=instance, purge=False)

    info = multipass.info(instance_name=instance)

    assert info["info"][instance]["state"] == "Deleted"


def test_delete_purge(instance, multipass):
    multipass.delete(instance_name=instance, purge=True)

    instances = multipass.list()

    assert instance not in instances


def test_exec(instance, multipass):
    proc = multipass.exec(
        instance_name=instance,
        command=["echo", "this is a test"],
        capture_output=True,
    )

    assert proc.stdout == b"this is a test\n"


def test_is_supported_version(multipass):
    assert multipass.is_supported_version() is True


def test_list(instance, multipass):
    instances = multipass.list()

    assert instance in instances


def test_mount_umount(instance, multipass, home_tmp_path):
    mount_target = f"{instance}:/tmp/mount-dir"
    test_file = home_tmp_path / "test.txt"
    test_file.write_text("this is a test")

    multipass.mount(
        source=home_tmp_path,
        target=mount_target,
    )

    proc = multipass.exec(
        command=["cat", "/tmp/mount-dir/test.txt"],
        instance_name=instance,
        capture_output=True,
        check=True,
    )

    assert proc.stdout == b"this is a test"

    multipass.umount(
        mount=mount_target,
    )

    with pytest.raises(subprocess.CalledProcessError):
        proc = multipass.exec(
            command=["test", "-f", "/tmp/mount-dir/test.txt"],
            instance_name=instance,
            check=True,
        )


def test_stop_start(instance, multipass):
    info = multipass.info(instance_name=instance)
    assert info["info"][instance]["state"] == "Running"

    multipass.stop(instance_name=instance)

    info = multipass.info(instance_name=instance)
    assert info["info"][instance]["state"] == "Stopped"

    multipass.start(instance_name=instance)

    info = multipass.info(instance_name=instance)
    assert info["info"][instance]["state"] == "Running"

    multipass.stop(instance_name=instance)

    info = multipass.info(instance_name=instance)
    assert info["info"][instance]["state"] == "Stopped"


def test_transfer_in(instance, multipass, home_tmp_path):
    test_file = home_tmp_path / "test.txt"
    test_file.write_text("this is a test")

    multipass.transfer(
        source=str(test_file),
        destination=f"{instance}:/tmp/foo",
    )

    proc = multipass.exec(
        command=["cat", "/tmp/foo"],
        instance_name=instance,
        capture_output=True,
        check=True,
    )

    assert proc.stdout == b"this is a test"


def test_transfer_out(instance, multipass, home_tmp_path):
    test_file = home_tmp_path / "test.txt"
    test_file.write_text("this is a test")

    multipass.transfer(
        source=str(test_file),
        destination=f"{instance}:/tmp/foo",
    )

    out_path = home_tmp_path / "out.txt"

    multipass.transfer(
        source=f"{instance}:/tmp/foo",
        destination=str(out_path),
    )

    assert out_path.read_text() == "this is a test"


def test_transfer_destination_io(instance, multipass, home_tmp_path):
    test_file = home_tmp_path / "test.txt"
    test_file.write_text("this is a test")

    multipass.transfer(
        source=str(test_file),
        destination=f"{instance}:/tmp/foo",
    )

    out_path = home_tmp_path / "out.txt"
    with out_path.open("wb") as stream:
        multipass.transfer_destination_io(
            source=f"{instance}:/tmp/foo",
            destination=stream,
        )

    assert out_path.read_text() == "this is a test"


@pytest.mark.slow()
def test_transfer_destination_io_large(instance, multipass, home_tmp_path):
    test_file = home_tmp_path / "test.txt"

    with test_file.open("wb") as stream:
        stream.seek(1024 * 1024 * 100)
        stream.write(b"test")

    assert test_file.stat().st_size == 104857604

    multipass.transfer(
        source=str(test_file),
        destination=f"{instance}:/tmp/foo",
    )

    out_path = home_tmp_path / "out.txt"
    with out_path.open("wb") as stream:
        multipass.transfer_destination_io(
            source=f"{instance}:/tmp/foo",
            destination=stream,
        )

    assert out_path.stat().st_size == 104857604


def test_transfer_source_io(instance, multipass):
    test_io = io.BytesIO(b"this is a test")

    multipass.transfer_source_io(
        source=test_io,
        destination=f"{instance}:/tmp/foo",
    )

    proc = multipass.exec(
        command=["cat", "/tmp/foo"],
        instance_name=instance,
        capture_output=True,
        check=True,
    )

    assert proc.stdout == b"this is a test"


@pytest.mark.slow()
def test_transfer_source_io_large(instance, multipass, home_tmp_path):
    test_file = home_tmp_path / "test.txt"

    with test_file.open("wb") as stream:
        stream.seek(1024 * 1024 * 100)
        stream.write(b"test")

    with test_file.open("rb") as stream:
        multipass.transfer_source_io(
            source=stream,
            destination=f"{instance}:/tmp/foo",
        )

    proc = multipass.exec(
        command=["du", "--bytes", "/tmp/foo"],
        instance_name=instance,
        capture_output=True,
        check=True,
    )

    assert proc.stdout == b"104857604\t/tmp/foo\n"


def test_wait_until_ready(multipass):
    multipass_version, multipassd_version = multipass.wait_until_ready()

    assert multipass_version is not None
    assert multipassd_version is not None


def test_version(multipass):
    multipass_version, multipassd_version = multipass.version()

    assert multipass_version is not None
    assert multipassd_version is not None



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Collection of bases used to configure build environments."""

# Backward compatible, will be removed in 2.0
import sys
from typing import Dict, NamedTuple, Tuple, Type, Union

from craft_providers.errors import BaseCompatibilityError, BaseConfigurationError
from craft_providers.base import Base

from . import almalinux, centos
from . import ubuntu
from . import ubuntu as buildd
from .ubuntu import BuilddBase, BuilddBaseAlias

sys.modules["craft_providers.bases.buildd"] = buildd

BaseAlias = Union[
    ubuntu.BuilddBaseAlias, almalinux.AlmaLinuxBaseAlias, centos.CentOSBaseAlias
]

__all__ = [
    "ubuntu",
    "centos",
    "BaseAlias",
    "BaseName",
    "BuilddBase",
    "BuilddBaseAlias",
    "BaseCompatibilityError",
    "BaseConfigurationError",
]


class BaseName(NamedTuple):
    """A base image, by distribution and version."""

    name: str
    version: str


BASE_NAME_TO_BASE_ALIAS: Dict[BaseName, BaseAlias] = {
    BaseName("ubuntu", "16.04"): ubuntu.BuilddBaseAlias.XENIAL,
    BaseName("ubuntu", "18.04"): ubuntu.BuilddBaseAlias.BIONIC,
    BaseName("ubuntu", "20.04"): ubuntu.BuilddBaseAlias.FOCAL,
    BaseName("ubuntu", "22.04"): ubuntu.BuilddBaseAlias.JAMMY,
    BaseName("ubuntu", "23.10"): ubuntu.BuilddBaseAlias.MANTIC,
    BaseName("ubuntu", "24.04"): ubuntu.BuilddBaseAlias.NOBLE,
    BaseName("ubuntu", "devel"): ubuntu.BuilddBaseAlias.DEVEL,
    BaseName("centos", "7"): centos.CentOSBaseAlias.SEVEN,
    BaseName("almalinux", "9"): almalinux.AlmaLinuxBaseAlias.NINE,
}


def get_base_alias(
    base_name: Tuple[str, str],
) -> BaseAlias:
    """Return a Base alias from a base (name, version) tuple."""
    base_name = BaseName(*base_name)
    if base_name.name == "ubuntu" and base_name in BASE_NAME_TO_BASE_ALIAS:
        return BASE_NAME_TO_BASE_ALIAS[base_name]

    # match other distributions sub-versions like 9.1 to 9
    _base_name = BaseName(base_name.name, base_name.version.split(".")[0])
    if _base_name in BASE_NAME_TO_BASE_ALIAS:
        return BASE_NAME_TO_BASE_ALIAS[_base_name]

    raise BaseConfigurationError(f"Base alias not found for {base_name}")


def get_base_from_alias(
    alias: BaseAlias,
) -> Type[Base]:
    """Return a Base class from a known base alias."""
    if isinstance(alias, ubuntu.BuilddBaseAlias):
        return ubuntu.BuilddBase

    if isinstance(alias, centos.CentOSBaseAlias):
        return centos.CentOSBase

    if isinstance(alias, almalinux.AlmaLinuxBaseAlias):
        return almalinux.AlmaLinuxBase

    raise BaseConfigurationError(f"Base not found for alias {alias}")
